"""
pytest fixtures and other good stuff

Author: Tom Fleet
Created: 05/02/2021
"""

import json
import pathlib
from typing import Any, Dict, List, Union

import pytest
import yaml


@pytest.fixture
def temp_config_file(tmp_path_factory):
    """
    Returns a totally valid config file.
    """

    config_file = tmp_path_factory.mktemp("temp").joinpath(".pytoil.yml")

    yaml_dict: Dict[str, str] = {
        "username": "tempfileuser",
        "token": "tempfiletoken",
        "projects_dir": "/Users/tempfileuser/projects",
        "vscode": True,
        "common_packages": ["black", "flake8", "mypy"],
    }

    with open(config_file, "w") as f:
        yaml.dump(yaml_dict, f)

    return config_file


@pytest.fixture
def temp_config_file_missing_key(tmp_path_factory):
    """
    Returns an otherwise-valid config file with a missing key.
    In this case, the `token` key.
    """

    config_file = tmp_path_factory.mktemp("temp").joinpath(".pytoil.yml")

    yaml_dict: Dict[str, str] = {
        "username": "tempfileuser",
        "projects_dir": "/Users/tempfileuser/projects",
        "vscode": True,
    }

    with open(config_file, "w") as f:
        yaml.dump(yaml_dict, f)

    return config_file


@pytest.fixture
def temp_config_file_key_with_blank_value(tmp_path_factory):
    """
    Returns an otherwise-valid config file but one of the keys
    has a blank value.

    In this case, the `token` key.

    A blank value is interpreted by pyyaml as None.
    """

    config_file = tmp_path_factory.mktemp("temp").joinpath(".pytoil.yml")

    yaml_dict: Dict[str, Union[str, None]] = {
        "username": "tempfileuser",
        "token": None,
        "projects_dir": "/Users/tempfileuser/projects",
        "vscode": True,
    }

    with open(config_file, "w") as f:
        yaml.dump(yaml_dict, f)

    return config_file


@pytest.fixture
def temp_config_file_misspelled_key(tmp_path_factory):
    """
    Returns an otherwise-valid config file with a misspelled key.
    In this case, the `token` key.
    """

    config_file = tmp_path_factory.mktemp("temp").joinpath(".pytoil.yml")

    yaml_dict: Dict[str, str] = {
        "username": "tempfileuser",
        "terrrken": "tempfiletoken",
        "projects_dir": "/Users/tempfileuser/projects",
    }

    with open(config_file, "w") as f:
        yaml.dump(yaml_dict, f)

    return config_file


@pytest.fixture
def temp_config_file_missing_username(tmp_path_factory):
    """
    Returns an otherwise-valid config file but the username key
    is blank

    A blank value is interpreted by pyyaml as None.
    """

    config_file = tmp_path_factory.mktemp("temp").joinpath(".pytoil.yml")

    yaml_dict: Dict[str, Union[str, None]] = {
        "username": None,
        "token": "tempfiletoken",
        "projects_dir": "/Users/tempfileuser/projects",
        "vscode": True,
    }

    with open(config_file, "w") as f:
        yaml.dump(yaml_dict, f)

    return config_file


@pytest.fixture
def temp_config_file_missing_token(tmp_path_factory):
    """
    Returns an otherwise-valid config file but the token key
    is blank

    A blank value is interpreted by pyyaml as None.
    """

    config_file = tmp_path_factory.mktemp("temp").joinpath(".pytoil.yml")

    yaml_dict: Dict[str, Union[str, None]] = {
        "username": "tempfileuser",
        "token": None,
        "projects_dir": "/Users/tempfileuser/projects",
        "vscode": True,
    }

    with open(config_file, "w") as f:
        yaml.dump(yaml_dict, f)

    return config_file


@pytest.fixture
def temp_config_file_missing_projects_dir(tmp_path_factory):
    """
    Returns an otherwise-valid config file but the projects_dir key
    is blank

    A blank value is interpreted by pyyaml as None.
    """

    config_file = tmp_path_factory.mktemp("temp").joinpath(".pytoil.yml")

    yaml_dict: Dict[str, Union[str, None]] = {
        "username": "tempfileuser",
        "token": "tempfiletoken",
        "projects_dir": None,
        "vscode": True,
    }

    with open(config_file, "w") as f:
        yaml.dump(yaml_dict, f)

    return config_file


@pytest.fixture
def temp_config_file_missing_vs_code(tmp_path_factory):
    """
    Returns an otherwise-valid config file but the projects_dir key
    is blank

    A blank value is interpreted by pyyaml as None.
    """

    config_file = tmp_path_factory.mktemp("temp").joinpath(".pytoil.yml")

    yaml_dict: Dict[str, Union[str, None]] = {
        "username": "tempfileuser",
        "token": "tempfiletoken",
        "projects_dir": "/Users/tempfileuser/projects",
        "vscode": None,
    }

    with open(config_file, "w") as f:
        yaml.dump(yaml_dict, f)

    return config_file


@pytest.fixture
def fake_api_response():

    response: Dict[str, Any] = [
        {"name": "repo1", "owner": "me", "blah": "bleh"},
        {"name": "repo2", "owner": "someguy", "blah": "bluh"},
        {"name": "repo3", "owner": "somegirl", "blah": "blah"},
    ]

    return response


@pytest.fixture
def temp_environment_yml(tmp_path_factory):
    """
    Returns a valid temporary environment.yml file
    """

    env_file: pathlib.Path = tmp_path_factory.mktemp("temp").joinpath("environment.yml")

    fake_env_info: Dict[str, Union[str, List[str]]] = {
        "name": "my_yml_env",
        "channels": ["defaults", "conda-forge"],
        "dependencies": [
            "python>3.7",
            "invoke",
            "black",
            "flake8",
            "isort",
            "mypy",
            "rich",
            "requests",
            "numpy",
            "pandas",
        ],
    }

    with open(env_file, "w") as f:
        yaml.dump(fake_env_info, f)

    return env_file


@pytest.fixture
def bad_temp_environment_yml(tmp_path_factory):
    """
    Returns an invalid temporary environment.yml file
    """

    env_file: pathlib.Path = tmp_path_factory.mktemp("temp").joinpath("environment.yml")

    # name must be a string
    fake_env_info = {
        "name": 300,
        "channels": ["defaults", "conda-forge"],
        "dependencies": [
            "python>3.7",
            "invoke",
            "black",
            "flake8",
            "isort",
            "mypy",
            "rich",
            "requests",
            "numpy",
            "pandas",
        ],
    }

    with open(env_file, "w") as f:
        yaml.dump(fake_env_info, f)

    return env_file


@pytest.fixture
def bad_temp_environment_yml_2(tmp_path_factory):
    """
    Returns an invalid temporary environment.yml file
    """

    env_file: pathlib.Path = tmp_path_factory.mktemp("temp").joinpath("environment.yml")

    # name must be a string
    fake_env_info = {
        "name": ["list", "of", "names"],
        "channels": ["defaults", "conda-forge"],
        "dependencies": [
            "python>3.7",
            "invoke",
            "black",
            "flake8",
            "isort",
            "mypy",
            "rich",
            "requests",
            "numpy",
            "pandas",
        ],
    }

    with open(env_file, "w") as f:
        yaml.dump(fake_env_info, f)

    return env_file


@pytest.fixture
def repo_folder_with_random_existing_files(tmp_path_factory):
    """
    Returns a temporary directory containing a few random files
    to test methods which check for a files existence within
    a certain directory.
    """

    folder: pathlib.Path = tmp_path_factory.mktemp("myrepo")

    files: List[str] = ["here.txt", "i_exist.yml", "hello.py", "me_too.json"]

    for file in files:
        # Create each file under the myrepo folder
        folder.joinpath(file).touch()

    return folder


@pytest.fixture
def fake_vscode_workspace_settings(tmp_path_factory):
    """
    Returns a temporary JSON file with fake
    vscode workspace settings inside.
    """

    fake_project_root: pathlib.Path = tmp_path_factory.mktemp("fakeproject")
    vscode_folder = fake_project_root.joinpath(".vscode")
    vscode_folder.mkdir(parents=True)

    settings_json = vscode_folder.joinpath("settings.json")
    settings_json.touch()

    settings_dict: Dict[str, Any] = {
        "editor.suggestSelection": True,
        "code-runner.runInTerminal": False,
        "python.linting.mypyEnabled": True,
        "python.linting.blackPath": "usr/bin/black",
        "randomSetting": "yes",
        "HowHardThisIsToComeUpWith": 10,
        "python.pythonPath": "/usr/bin/python",
        "python.testing.pytestEnabled": False,
    }

    with open(settings_json, mode="w", encoding="utf-8") as f:
        json.dump(settings_dict, f, sort_keys=True, indent=4)

    return settings_json


@pytest.fixture
def fake_home_folder_miniconda(tmp_path_factory):
    """
    Returns a faked $HOME folder. Should be used as the return
    value from pathlib.Path.home().

    Designed to test the auto detection of conda environment
    storage directory.
    """

    fake_home: pathlib.Path = tmp_path_factory.mktemp("home")

    # Add in the miniconda3/envs folder
    miniconda = fake_home.joinpath("miniconda3/envs")
    miniconda.mkdir(parents=True)

    return fake_home


@pytest.fixture
def fake_home_folder_anaconda(tmp_path_factory):
    """
    Returns a faked $HOME folder. Should be used as the return
    value from pathlib.Path.home().

    Designed to test the auto detection of conda environment
    storage directory.
    """

    fake_home: pathlib.Path = tmp_path_factory.mktemp("home")

    # Add in the anaconda3/envs folder
    anaconda = fake_home.joinpath("anaconda3/envs")
    anaconda.mkdir(parents=True)

    return fake_home


@pytest.fixture
def fake_home_folder_neither(tmp_path_factory):
    """
    Returns a faked $HOME but without any conda dirs.

    To test whether get_envs_dir raises correctly.
    """

    fake_home: pathlib.Path = tmp_path_factory.mktemp("home")

    return fake_home


@pytest.fixture
def fake_projects_dir(tmp_path_factory):
    """
    Returns a fake projects directory complete
    with subdirectories mocking development projects.
    """

    project_names: List[str] = [
        "project1",
        "myproject",
        "dingleproject",
        "anotherone",
        ".ishouldnt_show_up",
    ]

    projects_dir: pathlib.Path = tmp_path_factory.mktemp("projects")

    for project_name in project_names:
        project_path = projects_dir.joinpath(project_name)
        project_path.mkdir()

    return projects_dir


@pytest.fixture
def empty_projects_dir(tmp_path_factory):
    """
    Returns a fake projects directory with no projects inside it.
    """

    projects_dir: pathlib.Path = tmp_path_factory.mktemp("projects")

    return projects_dir



"""
Module managing VirtualEnvs

Author: Tom Fleet
Created: 07/03/2021
"""

import pathlib
import subprocess
from typing import List, Optional

import virtualenv

from pytoil.environments import BaseEnvironment
from pytoil.exceptions import MissingInterpreterError, VirtualenvAlreadyExistsError


class VirtualEnv(BaseEnvironment):
    def __init__(self, project_path: pathlib.Path) -> None:
        """
        Representation of a virtualenv.

        A VirtualEnv's `.executable` property (also a pathlib.Path) points
        to the python executable of the newly created virtualenv. This
        executable may or may not exist, and the existence check is
        the primary means of checking whether or not this virtualenv
        exists.

        Note: It is important not to resolve `.executable` as it could
        resolve back to the system python if it is a symlink.

        Args:
            project_path (pathlib.Path): The root path of the current project.
        """
        self._project_path = project_path.resolve()

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(project_path={self._project_path!r})"

    @property
    def project_path(self) -> pathlib.Path:
        return self._project_path

    @property
    def executable(self) -> pathlib.Path:
        return self._project_path.joinpath(".venv/bin/python")

    def exists(self) -> bool:
        """
        Checks whether the virtual environment exists by a proxy
        check if the `executable` exists.

        If this executable exists then both the project and virtual environment
        must also exist and must be valid.
        """
        return self.executable.exists()

    def create(self, packages: Optional[List[str]] = None) -> None:
        """
        Create a new virtualenv in `project_path`.

        Raises:
            VirtualenvAlreadyExistsError: If virtualenv already exists
                in `project_path`.
        """
        if self.exists():
            raise VirtualenvAlreadyExistsError(
                f"""Virtualenv with path: {self.executable}
                already exists"""
            )
        # Create a new virtualenv under the project, called ".venv"
        virtualenv.cli_run([f"{self.project_path.joinpath('.venv')}"])
        if packages:
            self.install(packages=packages)

    def update_seeds(self) -> None:
        """
        VirtualEnv will install 'seed' packages to a new
        environment: `pip`, `setuptools` and `wheel`.

        It is good practice to keep these packages fully up to date
        this method does exactly that by invoking pip from
        the virtualenvs executable.

        This is equivalent to running:
        `python -m pip install --upgrade pip setuptools wheel`
        from the command line with the virtualenv activated.
        """

        # Validate the executable
        if not self.exists():
            raise MissingInterpreterError(f"Interpreter: {self.executable} not found.")

        try:
            # Don't need to specify a 'cwd' because we have a resolved interpreter
            subprocess.run(
                [
                    f"{str(self.executable)}",
                    "-m",
                    "pip",
                    "install",
                    "--upgrade",
                    "pip",
                    "setuptools",
                    "wheel",
                ],
                check=True,
            )
        except subprocess.CalledProcessError:
            raise

    def install(self, packages: List[str]) -> None:
        """
        Generic `pip install` method.

        Takes a list of packages to install. All packages are passed through
        to pip so any versioning syntax will work as expected.

        Args:
            packages (List[str]): A list of valid packages to install.
            Analogous to simply calling `pip install *packages` from within
            the virtualenv. If only 1 package, still must be in a list e.g. `["black"]`.
        """

        # If we get here, method has been called correctly
        # Ensure seed packages are updated, also checks interpreter
        self.update_seeds()

        cmd: List[str] = [f"{self.executable}", "-m", "pip", "install"]

        cmd.extend(packages)

        # Run the constructed pip command
        try:
            subprocess.run(cmd, check=True)
        except subprocess.CalledProcessError:
            raise



"""
Tests for the CLI sync command.

Author: Tom Fleet
Created: 11/04/2021
"""

from typing import List

import pytest
from pytest_mock import MockerFixture
from typer.testing import CliRunner

from pytoil.cli.main import app
from pytoil.config import Config

runner = CliRunner()


def test_sync_all_does_nothing_if_local_and_remote_are_synced(
    mocker: MockerFixture, fake_projects_dir
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.sync.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # We use the local project names as repos so everything is in sync
    fake_repo_names: List[str] = [
        "project1",
        "myproject",
        "dingleproject",
        "anotherone",
    ]

    mocker.patch(
        "pytoil.api.api.API.get_repo_names",
        autospec=True,
        return_value=fake_repo_names,
    )

    # Make sure it doesn't actually clone anything
    mocker.patch("pytoil.repo.repo.Repo.clone", autospec=True, return_value=None)

    result = runner.invoke(app, ["sync", "all"])
    assert result.exit_code == 0

    assert (
        "All your remote repos already exist locally. Nothing to do!" in result.stdout
    )


def test_sync_all_respects_user_input_no(mocker: MockerFixture, fake_projects_dir):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.sync.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Now local and remotes are out of sync
    fake_repo_names: List[str] = [
        "repo1",
        "anotherrepo",
        "remote1",
        "project1",
        "myproject",
        "dingleproject",
        "anotherone",
    ]

    mocker.patch(
        "pytoil.api.api.API.get_repo_names",
        autospec=True,
        return_value=fake_repo_names,
    )

    # Make sure it doesn't actually clone anything
    mocker.patch("pytoil.repo.repo.Repo.clone", autospec=True, return_value=None)

    # No force, will prompt user. Here we input n
    # should abort and not do anything
    result = runner.invoke(app, ["sync", "all"], input="n")
    assert result.exit_code == 1

    assert "Aborted" in result.stdout


def test_sync_all_respects_user_input_yes(mocker: MockerFixture, fake_projects_dir):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.sync.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Now local and remotes are out of sync
    fake_repo_names: List[str] = [
        "repo1",
        "anotherrepo",
        "remote1",
        "project1",
        "myproject",
        "dingleproject",
        "anotherone",
    ]

    mocker.patch(
        "pytoil.api.api.API.get_repo_names",
        autospec=True,
        return_value=fake_repo_names,
    )

    # Make sure it doesn't actually clone anything
    mocker.patch("pytoil.repo.repo.Repo.clone", autospec=True, return_value=None)

    # No force, will prompt user. Here we input y
    # should now clone stuff
    result = runner.invoke(app, ["sync", "all"], input="y")
    assert result.exit_code == 0

    # Only first 3 are remote but not local
    assert "This will clone 3 repos" in result.stdout
    assert "Are you sure you want to proceed?" in result.stdout

    # Only first 3 are remote but not local
    for repo in fake_repo_names[:3]:
        assert f"Cloning: {repo!r}" in result.stdout


@pytest.mark.parametrize("force_flag", ["--force", "-f"])
def test_sync_all_respects_force_flag(
    mocker: MockerFixture, fake_projects_dir, force_flag
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.sync.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Now local and remotes are out of sync
    fake_repo_names: List[str] = [
        "repo1",
        "anotherrepo",
        "remote1",
        "project1",
        "myproject",
        "dingleproject",
        "anotherone",
    ]

    mocker.patch(
        "pytoil.api.api.API.get_repo_names",
        autospec=True,
        return_value=fake_repo_names,
    )

    # Make sure it doesn't actually clone anything
    mocker.patch("pytoil.repo.repo.Repo.clone", autospec=True, return_value=None)

    # Now we force, checking both flags work the same
    result = runner.invoke(app, ["sync", "all", force_flag])
    assert result.exit_code == 0

    # Only first 3 are remote but not local
    for repo in fake_repo_names[:3]:
        assert f"Cloning: {repo!r}" in result.stdout


def test_sync_these_does_nothing_if_local_and_remote_are_synced(
    mocker: MockerFixture, fake_projects_dir
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.sync.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # We use the local project names as repos so everything is in sync
    fake_repo_names: List[str] = [
        "project1",
        "myproject",
        "dingleproject",
        "anotherone",
    ]

    mocker.patch(
        "pytoil.api.api.API.get_repo_names",
        autospec=True,
        return_value=fake_repo_names,
    )

    # Make sure it doesn't actually clone anything
    mocker.patch("pytoil.repo.repo.Repo.clone", autospec=True, return_value=None)

    # We use a subset of the already synced projects as arguments
    # to these
    # Shouldnt do anything
    result = runner.invoke(
        app, ["sync", "these", "project1", "dingleproject", "anotherone"]
    )
    assert result.exit_code == 0

    assert (
        "All your remote repos already exist locally. Nothing to do." in result.stdout
    )


def test_sync_these_respects_user_input_no(mocker: MockerFixture, fake_projects_dir):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.sync.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Now local and remotes are out of sync
    fake_repo_names: List[str] = [
        "repo1",
        "anotherrepo",
        "remote1",
        "project1",
        "myproject",
        "dingleproject",
        "anotherone",
    ]

    mocker.patch(
        "pytoil.api.api.API.get_repo_names",
        autospec=True,
        return_value=fake_repo_names,
    )

    # Make sure it doesn't actually clone anything
    mocker.patch("pytoil.repo.repo.Repo.clone", autospec=True, return_value=None)

    # No force, will prompt user. Here we input n
    # should abort and not do anything
    result = runner.invoke(
        app, ["sync", "these", "repo1", "project1", "remote1"], input="n"
    )
    assert result.exit_code == 1

    assert "Aborted" in result.stdout


def test_sync_these_respects_user_input_yes(mocker: MockerFixture, fake_projects_dir):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.sync.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Now local and remotes are out of sync
    fake_repo_names: List[str] = [
        "repo1",
        "anotherrepo",
        "remote1",
        "project1",
        "myproject",
        "dingleproject",
        "anotherone",
    ]

    mocker.patch(
        "pytoil.api.api.API.get_repo_names",
        autospec=True,
        return_value=fake_repo_names,
    )

    # Make sure it doesn't actually clone anything
    mocker.patch("pytoil.repo.repo.Repo.clone", autospec=True, return_value=None)

    # No force, will prompt user. Here we input y
    # should only clone repo1, anotherrepo, and remote1, project1 is already local
    result = runner.invoke(
        app,
        ["sync", "these", "repo1", "project1", "anotherrepo", "remote1"],
        input="y",
    )
    assert result.exit_code == 0

    # Only first 3 are remote but not local
    assert "This will clone 3 repos" in result.stdout
    assert "Are you sure you want to proceed?" in result.stdout

    # Only first 3 are remote but not local
    for repo in fake_repo_names[:3]:
        assert f"Cloning: {repo!r}" in result.stdout


@pytest.mark.parametrize("force_flag", ["--force", "-f"])
def test_sync_these_respects_force_flag(
    mocker: MockerFixture, fake_projects_dir, force_flag
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.sync.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Now local and remotes are out of sync
    fake_repo_names: List[str] = [
        "repo1",
        "anotherrepo",
        "remote1",
        "project1",
        "myproject",
        "dingleproject",
        "anotherone",
    ]

    mocker.patch(
        "pytoil.api.api.API.get_repo_names",
        autospec=True,
        return_value=fake_repo_names,
    )

    # Make sure it doesn't actually clone anything
    mocker.patch("pytoil.repo.repo.Repo.clone", autospec=True, return_value=None)

    # Now we force, checking both flags work the same
    result = runner.invoke(
        app,
        [
            "sync",
            "these",
            "repo1",
            "project1",
            "anotherrepo",
            "remote1",
            force_flag,
        ],
    )
    assert result.exit_code == 0

    # Only first 3 are remote but not local
    for repo in fake_repo_names[:3]:
        assert f"Cloning: {repo!r}" in result.stdout



"""
Tests for the config module.

Author: Tom Fleet
Created: 05/02/2021
"""

import pathlib

import pytest
from pytest_mock import MockerFixture

import pytoil
from pytoil.config import DEFAULT_PROJECTS_DIR, Config
from pytoil.exceptions import InvalidConfigError


def test_config_init_default():

    config = Config()

    assert config.username == "UNSET"
    assert config.token == "UNSET"
    # Default development path
    assert config.projects_dir == DEFAULT_PROJECTS_DIR
    assert config.vscode is False


def test_config_init_passed():

    config = Config(
        username="me",
        token="definitelynotatoken",
        projects_dir=pathlib.Path("/Users/me/projects"),
        vscode=True,
    )

    assert config.username == "me"
    assert config.token == "definitelynotatoken"
    assert config.projects_dir == pathlib.Path("/Users/me/projects")
    assert config.vscode is True


def test_config_init_username():

    config = Config(username="me")

    assert config.username == "me"
    assert config.token == "UNSET"
    assert config.projects_dir == DEFAULT_PROJECTS_DIR
    assert config.vscode is False


def test_config_init_token():

    config = Config(token="definitelynotatoken")

    assert config.username == "UNSET"
    assert config.token == "definitelynotatoken"
    assert config.projects_dir == DEFAULT_PROJECTS_DIR
    assert config.vscode is False


def test_config_init_projects_dir():

    config = Config(projects_dir=pathlib.Path("/Users/madeup/path"))

    assert config.username == "UNSET"
    assert config.token == "UNSET"
    assert config.projects_dir == pathlib.Path("/Users/madeup/path")
    assert config.vscode is False


def test_config_init_vscode():

    config = Config(vscode=True)

    assert config.username == "UNSET"
    assert config.token == "UNSET"
    assert config.projects_dir == DEFAULT_PROJECTS_DIR
    assert config.vscode is True


def test_config_repr_default():

    config = Config()

    assert (
        config.__repr__()
        == "Config(username='UNSET', "
        + "token='UNSET', "
        + f"projects_dir={DEFAULT_PROJECTS_DIR!r}, "
        + "vscode=False, "
        + "common_packages=None)"
    )


def test_config_repr_passed():

    path = pathlib.Path("/Users/me/projects")

    config = Config(
        username="me",
        token="definitelynotatoken",
        projects_dir=path,
        vscode=True,
        common_packages=["black", "flake8", "mypy"],
    )

    expected = (
        "Config(username='me', "
        + "token='definitelynotatoken', "
        + f"projects_dir={path!r}, "
        + "vscode=True, "
        + "common_packages=['black', 'flake8', 'mypy'])"
    )

    assert config.__repr__() == expected


def test_config_dict_default():

    config = Config()

    assert config.__dict__ == {
        "username": "UNSET",
        "token": "UNSET",
        "projects_dir": DEFAULT_PROJECTS_DIR,
        "vscode": False,
        "common_packages": None,
    }


def test_config_dict_passed():

    path = pathlib.Path("/Users/me/projects")

    config = Config(
        username="me",
        token="definitelynotatoken",
        projects_dir=path,
        vscode=True,
        common_packages=["black", "flake8", "mypy"],
    )

    assert config.__dict__ == {
        "username": "me",
        "token": "definitelynotatoken",
        "projects_dir": path,
        "vscode": True,
        "common_packages": ["black", "flake8", "mypy"],
    }


def test_config_setters():

    config = Config()

    # Assert before
    assert config.username == "UNSET"
    assert config.token == "UNSET"
    assert config.projects_dir == DEFAULT_PROJECTS_DIR
    assert config.vscode is False

    # Change value using setters
    config.username = "me"
    config.token = "definitelynotatoken"
    config.projects_dir = pathlib.Path("/Users/me/projects")
    config.vscode = True

    # Assert after
    assert config.username == "me"
    assert config.token == "definitelynotatoken"
    assert config.projects_dir == pathlib.Path("/Users/me/projects")
    assert config.vscode is True


def test_config_get_good_file(temp_config_file, mocker: MockerFixture):
    """
    Tests config.get on a file with valid key value pairs.
    """
    # Patch out the default pointer to the config file for our temp fixture
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        config = Config.get()

        assert config.username == "tempfileuser"
        assert config.token == "tempfiletoken"
        assert config.projects_dir == pathlib.Path("/Users/tempfileuser/projects")
        assert config.vscode is True


def test_config_get_raises_on_missing_file(mocker: MockerFixture):
    """
    Tests config.get raises the correct exception if the config
    file is missing.
    """

    with mocker.patch.object(
        pytoil.config.config,
        "CONFIG_PATH",
        pathlib.Path("definitely/not/here/.pytoil.yml"),
    ):

        with pytest.raises(FileNotFoundError):
            Config.get()


@pytest.mark.parametrize(
    "name, token, projects_dir, vscode, common_packages, expected_dict",
    [
        (
            "me",
            "sillytoken",
            pathlib.Path("/Users/me/projects"),
            True,
            ["black", "flake8", "mypy"],
            {
                "username": "me",
                "token": "sillytoken",
                "projects_dir": "/Users/me/projects",
                "vscode": True,
                "common_packages": ["black", "flake8", "mypy"],
            },
        ),
        (
            "someguy",
            "loltoken",
            pathlib.Path("/Users/someguy/dingleprojects"),
            False,
            None,
            {
                "username": "someguy",
                "token": "loltoken",
                "projects_dir": "/Users/someguy/dingleprojects",
                "vscode": False,
                "common_packages": None,
            },
        ),
        (
            "dave",
            "hahahatoken",
            pathlib.Path("/Users/dave/hahaprojects"),
            True,
            ["pylint", "autopep8", "pytest"],
            {
                "username": "dave",
                "token": "hahahatoken",
                "projects_dir": "/Users/dave/hahaprojects",
                "vscode": True,
                "common_packages": ["pylint", "autopep8", "pytest"],
            },
        ),
    ],
)
def test_config_to_dict_returns_correct_values(
    name, token, projects_dir, vscode, common_packages, expected_dict
):

    config = Config(
        username=name,
        token=token,
        projects_dir=projects_dir,
        vscode=vscode,
        common_packages=common_packages,
    )

    assert config.to_dict() == expected_dict


def test_config_validate_raises_on_unset_username():

    config = Config(
        username="UNSET", token="definitelynotatoken", projects_dir="/Users/me/projects"
    )

    with pytest.raises(InvalidConfigError):
        config.validate()


def test_config_raise_if_unset_raises_on_unset_token():

    config = Config(username="me", token="UNSET", projects_dir="/Users/me/projects")

    with pytest.raises(InvalidConfigError):
        config.validate()


def test_config_show_outputs_correct_text(capsys):

    config = Config(
        username="me",
        token="UNSET",
        projects_dir=pathlib.Path("/Users/me/projects"),
        vscode=True,
        common_packages=["black", "flake8", "mypy"],
    )

    config.show()

    captured = capsys.readouterr()

    expected_output: str = (
        "username: 'me'\n"
        + "token: 'UNSET'\n"
        + "projects_dir: '/Users/me/projects'\n"
        + "vscode: True\n"
        + "common_packages: ['black', 'flake8', 'mypy']\n"
    )

    assert captured.out == expected_output


def test_config_write(mocker: MockerFixture, temp_config_file):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        original_config = Config.get()

        # Check config values before change
        assert original_config.username == "tempfileuser"
        assert original_config.token == "tempfiletoken"
        assert original_config.projects_dir == pathlib.Path(
            "/Users/tempfileuser/projects"
        )
        assert original_config.vscode is True

        config = Config(
            username="me",
            token="definitelynotatoken",
            projects_dir=pathlib.Path("/Users/me/projects"),
            vscode=False,
        )

        # Write these new attributes to the temp config file (overwriting)
        config.write()

        # Check config values after change

        new_config = Config.get()

        assert new_config.username == "me"
        assert new_config.token == "definitelynotatoken"
        assert new_config.projects_dir == pathlib.Path("/Users/me/projects")
        assert new_config.vscode is False



"""
CLI sync command.

Author: Tom Fleet
Created: 28/02/2021
"""

from typing import List, Set

import typer

from pytoil.api import API
from pytoil.cli.utils import get_local_project_set, get_remote_project_set
from pytoil.config import Config
from pytoil.repo import Repo

app = typer.Typer(no_args_is_help=True)


# Callback for documentation only.
@app.callback()
def sync() -> None:
    """
    Synchronise your local and remote projects.

    sync is a safe method in that existing local
    projects will not be modified in any way.
    """


# Can't call it all, keyword!
@app.command(name="all")
def all_(
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force sync without confirmation prompt.",
        show_default=False,
    )
) -> None:
    """
    Pull down all your remote projects that aren't already local.

    Will not operate on or modify in any way the projects that
    already exist.

    Examples:

    $ pytoil sync all

    $ pytoil sync all --force
    """

    # Everything below requires a valid config
    config = Config.get()
    config.validate()

    api = API()

    local_projects: Set[str] = get_local_project_set(config.projects_dir)

    remote_projects: Set[str] = get_remote_project_set(api)

    to_clone: Set[str] = remote_projects.difference(local_projects)

    n_to_clone: int = len(to_clone)

    # Internal correctness check
    assert n_to_clone >= 0

    if n_to_clone == 0:
        typer.secho(
            "\nAll your remote repos already exist locally. Nothing to do!",
            fg=typer.colors.GREEN,
        )
    else:
        if not force:
            typer.confirm(
                f"This will clone {n_to_clone} repos."
                + " Are you sure you want to proceed?",
                abort=True,
            )

        # User said force, so just go and do it
        for repo_name in to_clone:
            # Create the repo object and clone
            repo = Repo(name=repo_name)
            typer.echo(f"\nCloning: {repo.name!r}.")
            repo.clone()


@app.command()
def these(
    repos: List[str] = typer.Argument(..., help="List of remote repos to pull down."),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force sync without confirmation prompt.",
        show_default=False,
    ),
) -> None:
    """
    Only pull down specified remote projects.

    Any projects in the list that already exist locally
    will be skipped and will not be modified in any way.

    Examples:

    $ pytoil sync these project1 project2 project3

    $ pytoil sync these --force project1 project2
    """

    # Everything below requires a valid config
    config = Config.get()
    config.validate()

    local_projects: Set[str] = get_local_project_set(config.projects_dir)

    # Accept a list of repos from the user
    remote_projects: Set[str] = set(repos)

    to_clone: Set[str] = remote_projects.difference(local_projects)

    n_to_clone: int = len(to_clone)

    # Internal correctness check
    assert n_to_clone >= 0

    if n_to_clone == 0:
        typer.secho(
            "\nAll your remote repos already exist locally. Nothing to do.",
            fg=typer.colors.GREEN,
        )
    else:
        if not force:
            typer.confirm(
                f"This will clone {n_to_clone} repos."
                + " Are you sure you want to proceed?",
                abort=True,
            )

        # If user said no, typer will abort and the following will not run
        # If they said yes, it will run
        typer.echo("You said yes")
        for repo_name in to_clone:
            # Create the repo object and clone
            repo = Repo(name=repo_name)
            typer.echo(f"\nCloning: {repo.name!r}.")
            repo.clone()



"""
Tests for the CondaEnv management.

Author: Tom Fleet
Created: 10/02/2021
"""

import pathlib
import subprocess
from typing import List, NamedTuple

import pytest
from pytest_mock import MockerFixture

from pytoil.environments import CondaEnv
from pytoil.exceptions import (
    BadEnvironmentFileError,
    CondaNotInstalledError,
    UnknownCondaInstallationError,
    VirtualenvAlreadyExistsError,
    VirtualenvDoesNotExistError,
)


def test_condaenv_init():

    fake_project = pathlib.Path("/Users/me/project/fakeproject")

    env = CondaEnv(name="sillyenv", project_path=fake_project)

    assert env.name == "sillyenv"
    assert env.project_path == fake_project.resolve()


def test_condaenv_repr():

    fake_project = pathlib.Path("/Users/me/project/fakeproject")

    env = CondaEnv(name="sillyenv", project_path=fake_project)

    assert env.__repr__() == (
        f"CondaEnv(name='sillyenv', project_path={fake_project.resolve()!r})"
    )


def test_condaenv_raise_for_conda_raises_when_required(mocker: MockerFixture):

    fake_project = pathlib.Path("/Users/me/project/fakeproject")

    # Patch out shutil.which so it returns what we want it to
    mocker.patch(
        "pytoil.environments.conda.shutil.which", autospec=True, return_value=False
    )

    env = CondaEnv(name="sillyenv", project_path=fake_project)

    with pytest.raises(CondaNotInstalledError):
        env.raise_for_conda()


def test_condaenv_raise_for_conda_doesnt_raise_when_required(mocker: MockerFixture):

    fake_project = pathlib.Path("/Users/me/project/fakeproject")

    mocker.patch(
        "pytoil.environments.conda.shutil.which", autospec=True, return_value=True
    )

    env = CondaEnv(name="sillyenv", project_path=fake_project)

    # Test will fail if this raises
    env.raise_for_conda()


def test_condaenv_exists_returns_true_if_executable_exists(mocker: MockerFixture):

    fake_project = pathlib.Path("/Users/me/project/fakeproject")

    mocker.patch(
        "pytoil.environments.conda.pathlib.Path.exists",
        autospec=True,
        return_value=True,
    )

    mocker.patch(
        "pytoil.environments.conda.CondaEnv.get_envs_dir",
        autospec=True,
        return_value=pathlib.Path("/Users/me/miniconda3"),
    )

    env = CondaEnv(name="test", project_path=fake_project)

    assert env.exists() is True


def test_condaenv_exists_returns_false_if_executable_doesnt_exist(
    mocker: MockerFixture,
):

    fake_project = pathlib.Path("/Users/me/project/fakeproject")

    mocker.patch(
        "pytoil.environments.conda.pathlib.Path.exists",
        autospec=True,
        return_value=False,
    )

    mocker.patch(
        "pytoil.environments.conda.CondaEnv.get_envs_dir",
        autospec=True,
        return_value=pathlib.Path("/Users/me/miniconda3"),
    )

    env = CondaEnv(name="test", project_path=fake_project)

    assert env.exists() is False


@pytest.mark.parametrize(
    "name, packages, expected_cmd",
    [
        (
            "sillyenv",
            ["black", "mypy", "flake8"],
            [
                "conda",
                "create",
                "-y",
                "--name",
                "sillyenv",
                "python=3",
                "black",
                "mypy",
                "flake8",
            ],
        ),
        (
            "dingleenv",
            None,
            [
                "conda",
                "create",
                "-y",
                "--name",
                "dingleenv",
                "python=3",
            ],
        ),
    ],
)
def test_condaenv_create_is_correctly_called(
    mocker: MockerFixture, name, packages, expected_cmd
):

    fake_project = pathlib.Path("/Users/me/projects/fakeproject")

    # Trick the test into thinking the environment doesn't already exist
    mocker.patch(
        "pytoil.environments.CondaEnv.exists", autospec=True, return_value=False
    )

    mock_subprocess = mocker.patch(
        "pytoil.environments.conda.subprocess.run", autospec=True
    )

    env = CondaEnv(name=name, project_path=fake_project)

    env.create(packages=packages)

    mock_subprocess.assert_called_once_with(expected_cmd, check=True)


def test_condaenv_create_raises_if_already_exists(mocker: MockerFixture):

    fake_project = pathlib.Path("/Users/me/projects/fakeproject")

    # Trick the test into thinking the environment already exists
    mocker.patch(
        "pytoil.environments.CondaEnv.exists", autospec=True, return_value=True
    )

    with pytest.raises(VirtualenvAlreadyExistsError):
        env = CondaEnv(name="i_already_exist", project_path=fake_project)
        env.create()


def test_condaenv_create_raises_on_subprocess_error(mocker: MockerFixture):

    fake_project = pathlib.Path("/Users/me/projects/fakeproject")

    mocker.patch(
        "pytoil.environments.CondaEnv.exists", autospec=True, return_value=False
    )

    mocker.patch(
        "pytoil.environments.conda.subprocess.run",
        autospec=True,
        side_effect=[subprocess.CalledProcessError(-1, "cmd")],
    )

    with pytest.raises(subprocess.CalledProcessError):
        env = CondaEnv(name="error", project_path=fake_project)
        env.create()


def test_condaenv_create_from_yml_passes_correct_command(
    mocker: MockerFixture, temp_environment_yml
):

    # Make it think the environment doesn't already exist
    mocker.patch(
        "pytoil.environments.CondaEnv.exists", autospec=True, return_value=False
    )

    # Mock out the actual call to conda
    mock_subprocess = mocker.patch(
        "pytoil.environments.conda.subprocess.run", autospec=True
    )

    expected_cmd: List[str] = [
        "conda",
        "env",
        "create",
        "-y",
        "--file",
        f"{temp_environment_yml.resolve()}",
    ]

    env = CondaEnv.create_from_yml(project_path=temp_environment_yml.parent)

    mock_subprocess.assert_called_once_with(expected_cmd, check=True)

    # Assert the return is correct also
    assert env.name == "my_yml_env"
    assert env.project_path == temp_environment_yml.parent


def test_condaenv_create_from_yml_raises_on_missing_file(mocker: MockerFixture):

    fake_project = pathlib.Path("/Users/me/projects/fakeproject")

    mocker.patch(
        "pytoil.environments.CondaEnv.exists", autospec=True, return_value=False
    )

    with pytest.raises(FileNotFoundError):
        CondaEnv.create_from_yml(project_path=fake_project)


def test_condaenv_create_from_yml_raises_on_invalid_formatted_yml(
    mocker: MockerFixture, bad_temp_environment_yml, bad_temp_environment_yml_2
):

    mocker.patch(
        "pytoil.environments.CondaEnv.exists", autospec=True, return_value=False
    )

    with pytest.raises(BadEnvironmentFileError):
        CondaEnv.create_from_yml(project_path=bad_temp_environment_yml.parent)

    with pytest.raises(BadEnvironmentFileError):
        CondaEnv.create_from_yml(project_path=bad_temp_environment_yml_2.parent)


def test_condaenv_create_from_yml_raises_if_env_already_exists(
    mocker: MockerFixture, temp_environment_yml
):

    mocker.patch(
        "pytoil.environments.CondaEnv.exists", autospec=True, return_value=True
    )

    with pytest.raises(VirtualenvAlreadyExistsError):
        CondaEnv.create_from_yml(project_path=temp_environment_yml.parent)


def test_condaenv_create_from_yml_raises_on_subprocess_error(
    mocker: MockerFixture, temp_environment_yml
):

    mocker.patch(
        "pytoil.environments.CondaEnv.exists", autospec=True, return_value=False
    )

    mocker.patch(
        "pytoil.environments.conda.subprocess.run",
        autospec=True,
        side_effect=[subprocess.CalledProcessError(-1, "cmd")],
    )

    with pytest.raises(subprocess.CalledProcessError):
        CondaEnv.create_from_yml(project_path=temp_environment_yml.parent)


@pytest.mark.parametrize(
    "name, packages",
    [
        ("sillyenv", ["numpy", "pandas", "requests"]),
        ("dingleenv", ["black", "mypy", "dinglepython"]),
        ("blah", ["altair", "matplotlib", "seaborn"]),
    ],
)
def test_condaenv_install_passes_correct_command(mocker: MockerFixture, name, packages):

    fake_project = pathlib.Path("/Users/me/projects/fakeproject")

    mocker.patch(
        "pytoil.environments.CondaEnv.exists", autospec=True, return_value=True
    )

    mock_subprocess = mocker.patch(
        "pytoil.environments.conda.subprocess.run", autospec=True
    )

    expected_cmd: List[str] = [
        "conda",
        "install",
        "--name",
        name,
        "-y",
    ]

    expected_cmd.extend(packages)

    env = CondaEnv(name=name, project_path=fake_project)

    env.install(packages=packages)

    mock_subprocess.assert_called_once_with(expected_cmd, check=True)


def test_condaenv_install_raises_if_environment_doesnt_exist(mocker: MockerFixture):

    fake_project = pathlib.Path("/Users/me/projects/fakeproject")

    mocker.patch(
        "pytoil.environments.CondaEnv.exists", autospec=True, return_value=False
    )

    with pytest.raises(VirtualenvDoesNotExistError):
        env = CondaEnv(name="missing", project_path=fake_project)
        env.install(["black", "mypy", "pandas"])


def test_condaenv_install_raises_on_subprocess_error(mocker: MockerFixture):

    fake_project = pathlib.Path("/Users/me/projects/fakeproject")

    mocker.patch(
        "pytoil.environments.CondaEnv.exists", autospec=True, return_value=True
    )

    mocker.patch(
        "pytoil.environments.conda.subprocess.run",
        autospec=True,
        side_effect=[subprocess.CalledProcessError(-1, "cmd")],
    )

    with pytest.raises(subprocess.CalledProcessError):
        env = CondaEnv(name="error", project_path=fake_project)
        env.install(["pandas", "numpy"])


def test_condaenv_export_yml_raises_on_missing_env(mocker: MockerFixture):

    fake_project = pathlib.Path("/Users/me/projects/fakeproject")

    mocker.patch(
        "pytoil.environments.CondaEnv.exists", autospec=True, return_value=False
    )

    mocker.patch("pytoil.environments.conda.subprocess.run", autospec=True)

    with pytest.raises(VirtualenvDoesNotExistError):
        env = CondaEnv(name="condingle", project_path=fake_project)
        env.export_yml()


def test_condaenv_export_yml_raises_on_subprocess_error(mocker: MockerFixture):

    fake_project = pathlib.Path("/Users/me/projects/fakeproject")

    mocker.patch(
        "pytoil.environments.CondaEnv.exists", autospec=True, return_value=True
    )

    mocker.patch(
        "pytoil.environments.conda.subprocess.run",
        autospec=True,
        side_effect=[subprocess.CalledProcessError(-1, "cmd")],
    )

    with pytest.raises(subprocess.CalledProcessError):
        env = CondaEnv(name="blah", project_path=fake_project)
        env.export_yml()


def test_condaenv_export_yml_exports_correct_content(
    mocker: MockerFixture, temp_environment_yml
):

    mocker.patch(
        "pytoil.environments.CondaEnv.exists", autospec=True, return_value=True
    )

    class FakeStdOut(NamedTuple):
        # we need this so the returned object from subprocess.run
        # has a `.stdout` attribute
        stdout: str = """
        name: fake_conda_env
        channels:
        - defaults
        - conda-forge
        dependencies:
        - python=3
        - invoke
        - black
        - flake8
        - isort
        - mypy
        - rich
        - numpy
        - requests
        - pandas
        - matplotlib
        - seaborn
        - altair
        - altair_data_server
        - altair_saver
        - ipykernel
        - jupyterlab
        - sqlalchemy
        - pandas-profiling
        - hiplot
        """

    mock_subprocess = mocker.patch(
        "pytoil.environments.conda.subprocess.run",
        autospec=True,
        return_value=FakeStdOut(),
    )

    env = CondaEnv(name="fake_conda_env", project_path=temp_environment_yml.parent)

    env.export_yml()

    assert temp_environment_yml.read_text(encoding="utf-8") == FakeStdOut().stdout

    mock_subprocess.assert_called_once_with(
        ["conda", "env", "export", "--from-history", "--name", "fake_conda_env"],
        check=True,
        capture_output=True,
        encoding="utf-8",
    )


def test_condaenv_get_envs_dir_returns_correctly_for_miniconda(
    mocker: MockerFixture, fake_home_folder_miniconda: pathlib.Path
):

    mocker.patch(
        "pytoil.environments.conda.pathlib.Path.home",
        autospec=True,
        return_value=fake_home_folder_miniconda,
    )

    env = CondaEnv.get_envs_dir()

    expected_env_dir = fake_home_folder_miniconda.joinpath("miniconda3/envs")

    assert env == expected_env_dir


def test_condaenv_get_envs_dir_returns_correctly_for_anaconda(
    mocker: MockerFixture, fake_home_folder_anaconda: pathlib.Path
):

    mocker.patch(
        "pytoil.environments.conda.pathlib.Path.home",
        autospec=True,
        return_value=fake_home_folder_anaconda,
    )

    env = CondaEnv.get_envs_dir()

    expected_env_dir = fake_home_folder_anaconda.joinpath("anaconda3/envs")

    assert env == expected_env_dir


def test_condaenv_get_envs_dir_raises_if_neither_found(
    mocker: MockerFixture, fake_home_folder_neither: pathlib.Path
):

    mocker.patch(
        "pytoil.environments.conda.pathlib.Path.home",
        autospec=True,
        return_value=fake_home_folder_neither,
    )

    with pytest.raises(UnknownCondaInstallationError):
        CondaEnv.get_envs_dir()



"""
Tests for the pytoil new command.

Author: Tom Fleet
Created: 11/04/2021
"""

import pytest
from pytest_mock import MockerFixture
from typer.testing import CliRunner

from pytoil.cli.main import app
from pytoil.config import Config

runner = CliRunner()


def test_new_suggests_checkout_if_already_exists_local(
    mocker: MockerFixture, fake_projects_dir
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.main.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Try to create one that is already in fake_projects_dir
    myproject_path = fake_projects_dir.joinpath("myproject")
    result = runner.invoke(app, ["new", "myproject"])
    assert result.exit_code == 1

    assert (
        f"Project: 'myproject' already exists locally at '{myproject_path}'."
        in result.stdout
    )

    assert "To resume an existing project, use 'checkout'." in result.stdout
    assert "Example: '$ pytoil checkout myproject'." in result.stdout
    assert "Aborted!" in result.stdout


def test_new_suggests_checkout_if_already_exists_remote(
    mocker: MockerFixture, fake_projects_dir
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.main.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Whatever we try and create now, it will think it exists on GitHub
    mocker.patch("pytoil.cli.main.Repo.exists_remote", autospec=True, return_value=True)

    # Try to create one that is already in fake_projects_dir
    result = runner.invoke(app, ["new", "remote1"])
    assert result.exit_code == 1

    assert "Project: 'remote1' already exists on GitHub." in result.stdout

    assert "To resume an existing project, use 'checkout'." in result.stdout
    assert "Example: '$ pytoil checkout remote1'." in result.stdout
    assert "Aborted!" in result.stdout


@pytest.mark.parametrize(
    "cookie_option",
    ["--cookie", "-c"],
)
def test_new_with_cookiecutter_correctly_invokes_cookiecutter(
    mocker: MockerFixture, fake_projects_dir, cookie_option
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=False,
    )

    mocker.patch(
        "pytoil.cli.main.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Whatever we try and create now, it will think it doesn't already exist
    mocker.patch(
        "pytoil.cli.main.Repo.exists_remote", autospec=True, return_value=False
    )
    mocker.patch("pytoil.cli.main.Repo.exists_local", autospec=True, return_value=False)

    # Make sure it doesn't actually do anything
    mock_cookie = mocker.patch("pytoil.cli.main.cookiecutter", autospec=True)

    result = runner.invoke(
        app,
        [
            "new",
            "myproject",
            cookie_option,
            "https://github.com/some/cookie.git",
        ],
    )

    assert result.exit_code == 0

    assert (
        "Creating project: 'myproject' with cookiecutter template: "
        + "'https://github.com/some/cookie.git'."
        in result.stdout
    )

    mock_cookie.assert_called_once_with(
        template="https://github.com/some/cookie.git", output_dir=fake_projects_dir
    )


@pytest.mark.parametrize(
    "venv_option, venv_backend",
    [
        ("--venv", "virtualenv"),
        ("-v", "virtualenv"),
        ("--venv", "conda"),
        ("-v", "conda"),
    ],
)
def test_new_with_venv_correctly_creates_an_environment(
    mocker: MockerFixture,
    fake_projects_dir,
    venv_option,
    venv_backend,
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=False,
    )

    mocker.patch(
        "pytoil.cli.main.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Whatever we try and create now, it will think it doesn't already exist
    mocker.patch(
        "pytoil.cli.main.Repo.exists_remote", autospec=True, return_value=False
    )
    mocker.patch("pytoil.cli.main.Repo.exists_local", autospec=True, return_value=False)

    # Make it think we have a valid conda
    # And a valid virtualenv
    mocker.patch(
        "pytoil.cli.main.CondaEnv.exists",
        autospec=True,
        return_value=False,
    )

    mocker.patch(
        "pytoil.cli.main.VirtualEnv.exists",
        autospec=True,
        return_value=False,
    )

    # Make sure it doesn't actually do anything
    mock_virtualenv_create = mocker.patch(
        "pytoil.cli.main.VirtualEnv.create", autospec=True
    )
    mocker.patch("pytoil.cli.main.VirtualEnv.update_seeds", autospec=True)

    mock_conda_create = mocker.patch("pytoil.cli.main.CondaEnv.create", autospec=True)

    result = runner.invoke(
        app,
        [
            "new",
            "mynewproject",
            venv_option,
            venv_backend,
        ],
    )

    assert result.exit_code == 0

    if venv_backend == "virtualenv":
        assert "Creating virtualenv for 'mynewproject'." in result.stdout
        mock_virtualenv_create.assert_called_once()

    if venv_backend == "conda":
        assert "Creating conda environment for 'mynewproject'." in result.stdout
        mock_conda_create.assert_called_once()


def test_new_with_no_venv_doesnt_make_a_virtual_environment(
    mocker: MockerFixture, fake_projects_dir
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=False,
    )

    mocker.patch(
        "pytoil.cli.main.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Whatever we try and create now, it will think it doesn't already exist
    mocker.patch(
        "pytoil.cli.main.Repo.exists_remote", autospec=True, return_value=False
    )
    mocker.patch("pytoil.cli.main.Repo.exists_local", autospec=True, return_value=False)

    result = runner.invoke(app, ["new", "mynewproject"])
    assert result.exit_code == 0

    assert (
        "Virtual environment not requested. Skipping environment creation."
        in result.stdout
    )


def test_new_no_virtualenv_still_opens_code(mocker: MockerFixture, fake_projects_dir):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.main.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Whatever we try and create now, it will think it doesn't already exist
    mocker.patch(
        "pytoil.cli.main.Repo.exists_remote", autospec=True, return_value=False
    )
    mocker.patch("pytoil.cli.main.Repo.exists_local", autospec=True, return_value=False)

    mock_code_open = mocker.patch("pytoil.cli.main.VSCode.open", autospec=True)

    result = runner.invoke(app, ["new", "mynewproject"])

    assert result.exit_code == 0
    assert (
        "Virtual environment not requested. Skipping environment creation."
        in result.stdout
    )
    mock_code_open.assert_called_once()


@pytest.mark.parametrize("venv_option", ["--venv", "-v"])
def test_new_with_virtualenv_sets_pythonpath_and_opens_code(
    mocker: MockerFixture, fake_projects_dir, venv_option
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.main.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Whatever we try and create now, it will think it doesn't already exist
    mocker.patch(
        "pytoil.cli.main.Repo.exists_remote", autospec=True, return_value=False
    )
    mocker.patch("pytoil.cli.main.Repo.exists_local", autospec=True, return_value=False)

    mocker.patch(
        "pytoil.cli.main.VirtualEnv.exists",
        autospec=True,
        return_value=False,
    )

    # Make sure it doesn't actually do anything
    mock_virtualenv_create = mocker.patch(
        "pytoil.cli.main.VirtualEnv.create", autospec=True
    )
    mocker.patch("pytoil.cli.main.VirtualEnv.update_seeds", autospec=True)

    mock_code_open = mocker.patch("pytoil.cli.main.VSCode.open", autospec=True)
    mock_code_ppath = mocker.patch(
        "pytoil.cli.main.VSCode.set_python_path", autospec=True
    )

    result = runner.invoke(app, ["new", "mynewproject", venv_option, "virtualenv"])

    assert result.exit_code == 0

    assert "Creating virtualenv for 'mynewproject'." in result.stdout

    mock_virtualenv_create.assert_called_once()

    assert "Setting 'python.pythonPath' in VSCode workspace." in result.stdout
    mock_code_ppath.assert_called_once()

    assert "Opening 'mynewproject' in VSCode..." in result.stdout
    mock_code_open.assert_called_once()


@pytest.mark.parametrize("venv_option", ["--venv", "-v"])
def test_new_with_conda_sets_pythonpath_and_opens_code(
    mocker: MockerFixture, fake_projects_dir, venv_option
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.main.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Whatever we try and create now, it will think it doesn't already exist
    mocker.patch(
        "pytoil.cli.main.Repo.exists_remote", autospec=True, return_value=False
    )
    mocker.patch("pytoil.cli.main.Repo.exists_local", autospec=True, return_value=False)

    mocker.patch(
        "pytoil.cli.main.CondaEnv.exists",
        autospec=True,
        return_value=False,
    )

    mocker.patch(
        "pytoil.cli.main.CondaEnv.get_envs_dir",
        autospec=True,
        return_value=fake_projects_dir.parent.joinpath("miniconda3"),
    )

    # Make sure it doesn't actually do anything
    mock_conda_create = mocker.patch("pytoil.cli.main.CondaEnv.create", autospec=True)

    mock_code_open = mocker.patch("pytoil.cli.main.VSCode.open", autospec=True)
    mock_code_ppath = mocker.patch(
        "pytoil.cli.main.VSCode.set_python_path", autospec=True
    )

    result = runner.invoke(app, ["new", "mynewproject", venv_option, "conda"])

    assert result.exit_code == 0

    assert "Creating conda environment for 'mynewproject'." in result.stdout

    mock_conda_create.assert_called_once()

    assert "Setting 'python.pythonPath' in VSCode workspace." in result.stdout
    mock_code_ppath.assert_called_once()

    assert "Opening 'mynewproject' in VSCode..." in result.stdout
    mock_code_open.assert_called_once()



"""
Test for version callback.

Author: Tom Fleet
Created: 13/04/2021
"""

import pytest
from typer.testing import CliRunner

from pytoil import __version__
from pytoil.cli.main import app

runner = CliRunner()


@pytest.mark.parametrize("version_flag", ["--version", "-V"])
def test_version_flag_correctly_displays_actual_version(version_flag):

    result = runner.invoke(app, [version_flag])
    assert result.exit_code == 0
    assert f"pytoil version: {__version__}" in result.stdout



"""
Tests for the repo module.

Author: Tom Fleet
Created: 05/02/2021
"""


import pathlib
import subprocess
from typing import NamedTuple

import pytest
from pytest_mock import MockerFixture

import pytoil
from pytoil.environments import CondaEnv, VirtualEnv
from pytoil.exceptions import (
    GitNotInstalledError,
    InvalidRepoPathError,
    InvalidURLError,
    LocalRepoExistsError,
    RepoNotFoundError,
)
from pytoil.repo import Repo


def test_repo_init(mocker: MockerFixture, temp_config_file):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        repo = Repo(owner="me", name="myproject")

        assert repo.owner == "me"
        assert repo.name == "myproject"
        assert repo.url == "https://github.com/me/myproject.git"
        assert repo.path == pathlib.Path("/Users/tempfileuser/projects/myproject")


def test_repo_init_defaults(mocker: MockerFixture, temp_config_file):

    # Patch out the config path to be our temp file
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # name is required
        repo = Repo(name="diffproject")

        assert repo.owner == "tempfileuser"
        assert repo.name == "diffproject"
        assert repo.url == "https://github.com/tempfileuser/diffproject.git"
        assert repo.path == pathlib.Path("/Users/tempfileuser/projects/diffproject")


def test_repo_repr(mocker: MockerFixture, temp_config_file):

    # Patch out to our fake config file to make sure it grabs from the config
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        repo = Repo(owner="me", name="myproject")

        assert repo.__repr__() == "Repo(owner='me', name='myproject')"


@pytest.mark.parametrize(
    "url, owner, name",
    [
        ("https://github.com/dingleuser/dinglerepo.git", "dingleuser", "dinglerepo"),
        (
            "https://github.com/MySuperUser/s1llypr0ject.git",
            "MySuperUser",
            "s1llypr0ject",
        ),
        (
            "https://github.com/FollowTheProcess/pytoil.git",
            "FollowTheProcess",
            "pytoil",
        ),
        ("https://github.com/W31rdUsern4m3/blah.git", "W31rdUsern4m3", "blah"),
        ("https://github.com/HelloDave/dave.git", "HelloDave", "dave"),
    ],
)
def test_repo_from_url(mocker: MockerFixture, temp_config_file, url, owner, name):

    # Patch out to our fake config file to make sure it grabs from the config
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        repo = Repo.from_url(url=url)

        assert repo.owner == owner
        assert repo.name == name

        # Make sure it reconstructs the url properly
        assert repo.url == url
        assert repo.path == pathlib.Path(f"/Users/tempfileuser/projects/{name}")


@pytest.mark.parametrize(
    "url",
    [
        "https://nothub.com/me/project.git",
        "",
        "http://github.com/me/project.git",
        "https://github.com/What5wk91yn-msbnu-t/what.git",
        "https://github.com/:::punctuation[]';]'[-=]/project.git"
        "https://github.com/doesntend/indotgit",
    ],
)
def test_repo_from_url_raises_on_bad_url(mocker: MockerFixture, temp_config_file, url):

    # Patch out to our fake config file to make sure it grabs from the config
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        with pytest.raises(InvalidURLError):
            Repo.from_url(url=url)


@pytest.mark.parametrize(
    "path, owner, name, url",
    [
        (
            "dingleuser/dinglerepo",
            "dingleuser",
            "dinglerepo",
            "https://github.com/dingleuser/dinglerepo.git",
        ),
        (
            "MySuperUser/s1llypr0ject",
            "MySuperUser",
            "s1llypr0ject",
            "https://github.com/MySuperUser/s1llypr0ject.git",
        ),
        (
            "FollowTheProcess/pytoil",
            "FollowTheProcess",
            "pytoil",
            "https://github.com/FollowTheProcess/pytoil.git",
        ),
        (
            "W31rdUsern4m3/blah",
            "W31rdUsern4m3",
            "blah",
            "https://github.com/W31rdUsern4m3/blah.git",
        ),
        (
            "HelloDave/dave",
            "HelloDave",
            "dave",
            "https://github.com/HelloDave/dave.git",
        ),
    ],
)
def test_repo_from_path(
    mocker: MockerFixture, temp_config_file, path, owner, name, url
):

    # Patch out to our fake config file to make sure it grabs from the config
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        repo = Repo.from_path(path=path)

        assert repo.owner == owner
        assert repo.name == name

        # Make sure it reconstructs the url properly
        assert repo.url == url
        assert repo.path == pathlib.Path(f"/Users/tempfileuser/projects/{name}")


@pytest.mark.parametrize(
    "path",
    ["", "What5wk91yn-msbnu-t/what", ":::punctuation[]';]'[-=]/project"],
)
def test_repo_from_path_raises_on_bad_path(
    mocker: MockerFixture, temp_config_file, path
):

    # Patch out to our fake config file to make sure it grabs from the config
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        with pytest.raises(InvalidRepoPathError):
            Repo.from_path(path=path)


def test_repo_exists_local_returns_true_if_path_exists(
    mocker: MockerFixture, temp_config_file
):

    # Patch out the config file to point to our temporary one
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Patch out Repo.path to something we know exists: this file
        with mocker.patch.object(pytoil.repo.Repo, "path", pathlib.Path(__file__)):

            repo = Repo(name="fakerepo")

            assert repo.exists_local() is True


def test_repo_exists_local_returns_false_if_path_doesnt_exist(
    mocker: MockerFixture, temp_config_file
):

    # Patch out the config file to point to our temporary one
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Patch out Repo.path to something we know doesn't exist
        with mocker.patch.object(pytoil.repo.Repo, "path", pathlib.Path("not/here")):

            repo = Repo(name="fakerepo")

            assert repo.exists_local() is False


@pytest.mark.parametrize(
    "non_existing_repo_name", ["noexist1", "noexist2", "noexist3", "noexist4"]
)
def test_repo_exists_remote_returns_false_on_missing_repo(
    mocker: MockerFixture, temp_config_file, non_existing_repo_name
):

    # Patch out the config file to point to our temporary one
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Patch out api.get to always raise a 404 not found
        # which is our indication in `exists_remote` that the repo doesn't exist
        mocker.patch(
            "pytoil.api.API.get_repo_names",
            autospec=True,
            return_value={"exist1", "exist2", "exist3", "exist4"},
        )

        # Rest of the params will be filled in by our patched config file
        repo = Repo(name=non_existing_repo_name)

        assert repo.exists_remote() is False


@pytest.mark.parametrize("existing_repo_name", ["exist1", "exist2", "exist3", "exist4"])
def test_repo_exists_remote_returns_true_on_valid_repo(
    mocker: MockerFixture, temp_config_file, existing_repo_name
):

    # Same trick with the config file
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Now patch out API.get_repo to return some arbitrary dict
        # Indication that everything worked okay
        mocker.patch(
            "pytoil.api.API.get_repo_names",
            autospec=True,
            return_value={"exist1", "exist2", "exist3", "exist4"},
        )

        repo = Repo(name=existing_repo_name)

        assert repo.exists_remote() is True


@pytest.mark.parametrize("which_return", ["", None, False])
def test_repo_clone_raises_on_invalid_git(
    mocker: MockerFixture, temp_config_file, which_return
):

    # Same trick with the config file
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Patch out shutil.which
        mocker.patch(
            "pytoil.repo.repo.shutil.which", autospec=True, return_value=which_return
        )

        with pytest.raises(GitNotInstalledError):
            repo = Repo(owner="me", name="myproject")
            repo.clone()


@pytest.mark.parametrize("which_return", ["", None, False])
def test_repo_init_raises_on_invalid_git(
    mocker: MockerFixture, temp_config_file, which_return
):

    # Same trick with the config file
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Patch out shutil.which
        mocker.patch(
            "pytoil.repo.repo.shutil.which", autospec=True, return_value=which_return
        )

        with pytest.raises(GitNotInstalledError):
            repo = Repo(owner="me", name="myproject")
            repo.init()


def test_repo_clone_raises_if_local_repo_already_exists(
    mocker: MockerFixture, temp_config_file
):

    # Same trick with the config file
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Make it look like we have a valid git
        mocker.patch("pytoil.repo.repo.shutil.which", autospec=True, return_value=True)

        # Make it think the repo already exists locally
        mocker.patch("pytoil.repo.Repo.exists_local", autospec=True, return_value=True)

        with pytest.raises(LocalRepoExistsError):
            repo = Repo(owner="me", name="myproject")
            repo.clone()


def test_repo_clone_correctly_calls_git(mocker: MockerFixture, temp_config_file):

    # Same trick with the config file
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Make it look like we have a valid git
        mocker.patch("pytoil.repo.repo.shutil.which", autospec=True, return_value=True)

        # Ensure the repo doesnt already exist
        mocker.patch("pytoil.repo.Repo.exists_local", autospec=True, return_value=False)

        # Ensure the repo "exists" on github
        mocker.patch("pytoil.repo.Repo.exists_remote", autospec=True, return_value=True)

        # Mock the subprocess of calling git
        mock_subprocess = mocker.patch("pytoil.repo.repo.subprocess.run", autospec=True)

        repo = Repo(name="fakerepo")

        repo.clone()

        # Assert git would have been called with correct args
        mock_subprocess.assert_called_once_with(
            ["git", "clone", "https://github.com/tempfileuser/fakerepo.git"],
            cwd=pathlib.Path("/Users/tempfileuser/projects"),
            check=True,
        )

        # Assert the path was updated
        assert repo.path == pathlib.Path("/Users/tempfileuser/projects/fakerepo")


def test_repo_init_correctly_calls_git(mocker: MockerFixture, temp_config_file):

    # Same trick with the config file
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Make it look like we have a valid git
        mocker.patch("pytoil.repo.repo.shutil.which", autospec=True, return_value=True)

        # Ensure the repo doesnt already exist
        mocker.patch("pytoil.repo.Repo.exists_local", autospec=True, return_value=False)

        # Mock the subprocess of calling git
        mock_subprocess = mocker.patch("pytoil.repo.repo.subprocess.run", autospec=True)

        repo = Repo(name="fakerepo")

        repo.init()

        # Assert git would have been called with correct args
        mock_subprocess.assert_called_once_with(
            ["git", "init"],
            cwd=pathlib.Path("/Users/tempfileuser/projects/fakerepo"),
            check=True,
        )


def test_repo_clone_raises_subprocess_error_if_anything_goes_wrong(
    mocker: MockerFixture, temp_config_file
):

    # Same trick with the config file
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Make it look like we have a valid git
        mocker.patch("pytoil.repo.repo.shutil.which", autospec=True, return_value=True)

        # Ensure the repo doesnt already exist
        mocker.patch("pytoil.repo.Repo.exists_local", autospec=True, return_value=False)

        # Ensure the repo "exists" on github
        mocker.patch("pytoil.repo.Repo.exists_remote", autospec=True, return_value=True)

        # Mock the subprocess of calling git, but have it raise an error
        mock_subprocess = mocker.patch(
            "pytoil.repo.repo.subprocess.run",
            autospec=True,
            side_effect=[subprocess.CalledProcessError(-1, "cmd")],
        )

        repo = Repo(name="fakerepo")

        with pytest.raises(subprocess.CalledProcessError):

            repo.clone()

            # Assert git would have been called with correct args
            mock_subprocess.assert_called_once_with(
                ["git", "clone", "https://github.com/tempfileuser/fakerepo.git"],
                cwd=pathlib.Path("/Users/tempfileuser/projects"),
                check=True,
            )


def test_repo_init_raises_subprocess_error_if_anything_goes_wrong(
    mocker: MockerFixture, temp_config_file
):

    # Same trick with the config file
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Make it look like we have a valid git
        mocker.patch("pytoil.repo.repo.shutil.which", autospec=True, return_value=True)

        # Ensure the repo doesnt already exist
        mocker.patch("pytoil.repo.Repo.exists_local", autospec=True, return_value=False)

        # Mock the subprocess of calling git, but have it raise an error
        mock_subprocess = mocker.patch(
            "pytoil.repo.repo.subprocess.run",
            autospec=True,
            side_effect=[subprocess.CalledProcessError(-1, "cmd")],
        )

        repo = Repo(name="fakerepo")

        with pytest.raises(subprocess.CalledProcessError):

            repo.init()

            # Assert git would have been called with correct args
            mock_subprocess.assert_called_once_with(
                ["git", "clone"],
                cwd=pathlib.Path("/Users/tempfileuser/projects/fakerepo"),
                check=True,
            )


def test_repo_clone_raises_on_missing_remote_repo(
    mocker: MockerFixture, temp_config_file
):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Make it look like we have a valid git
        mocker.patch("pytoil.repo.repo.shutil.which", autospec=True, return_value=True)

        # Ensure the repo doesnt already exist
        mocker.patch("pytoil.repo.Repo.exists_local", autospec=True, return_value=False)

        # Ensure the repo doesnt already exist
        mocker.patch(
            "pytoil.repo.Repo.exists_remote", autospec=True, return_value=False
        )

        repo = Repo(name="fakerepo")

        with pytest.raises(RepoNotFoundError):
            repo.clone()


@pytest.mark.parametrize(
    "file, exists",
    [
        ("here.txt", True),
        ("i_exist.yml", True),
        ("hello.py", True),
        ("me_too.json", True),
        ("not_me_though.csv", False),
        ("me_neither.toml", False),
        ("i_never_existed.cfg", False),
        ("what_file.ini", False),
    ],
)
def test_does_file_exist(
    mocker: MockerFixture,
    temp_config_file,
    repo_folder_with_random_existing_files,
    file,
    exists,
):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Trick it into thinking a repo exists
        mocker.patch("pytoil.repo.Repo.exists_local", autospec=True, return_value=True)

        folder: pathlib.Path = repo_folder_with_random_existing_files

        repo = Repo(name="myrepo")

        # Set the repo path to point to our folder
        repo.path = folder

        # Test the file exists method
        assert repo._does_file_exist(file) is exists


def test_does_file_exist_raises_on_non_existent_repo(
    mocker: MockerFixture, temp_config_file, repo_folder_with_random_existing_files
):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Return False for repo.exists_local
        mocker.patch("pytoil.repo.Repo.exists_local", autospec=True, return_value=False)

        folder: pathlib.Path = repo_folder_with_random_existing_files

        repo = Repo(name="myrepo")

        # Set the repo path to point to our folder
        repo.path = folder

        with pytest.raises(RepoNotFoundError):
            repo._does_file_exist("here.txt")


def test_is_setuptools(
    mocker: MockerFixture, temp_config_file, repo_folder_with_random_existing_files
):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Return True for repo.exists_local
        mocker.patch("pytoil.repo.Repo.exists_local", autospec=True, return_value=True)

        folder: pathlib.Path = repo_folder_with_random_existing_files

        repo = Repo(name="myrepo")

        # Set the repo path to point to our folder
        repo.path = folder

        # Add in a setup.py and setup.cfg
        folder.joinpath("setup.py").touch()
        folder.joinpath("setup.cfg").touch()

        # Should now return True
        assert repo.is_setuptools() is True

        # Now remove setup.py
        folder.joinpath("setup.py").unlink()

        # Should still return True
        assert repo.is_setuptools() is True

        # Now remove setup.cfg
        folder.joinpath("setup.cfg").unlink()

        # Should now return False
        assert repo.is_setuptools() is False

        # Now just a setup.py
        folder.joinpath("setup.py").touch()

        # Should again return True
        assert repo.is_setuptools() is True


def test_is_editable(
    mocker: MockerFixture, temp_config_file, repo_folder_with_random_existing_files
):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Return True for repo.exists_local
        mocker.patch("pytoil.repo.Repo.exists_local", autospec=True, return_value=True)

        folder: pathlib.Path = repo_folder_with_random_existing_files

        repo = Repo(name="myrepo")

        # Set the repo path to point to our folder
        repo.path = folder

        # Add in a setup.py
        folder.joinpath("setup.py").touch()

        # Should now return True
        assert repo.is_editable() is True

        # Now remove setup.py
        folder.joinpath("setup.py").unlink()

        # Should now return False
        assert repo.is_setuptools() is False


def test_is_conda(
    mocker: MockerFixture, temp_config_file, repo_folder_with_random_existing_files
):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Return True for repo.exists_local
        mocker.patch("pytoil.repo.Repo.exists_local", autospec=True, return_value=True)

        folder: pathlib.Path = repo_folder_with_random_existing_files

        repo = Repo(name="myrepo")

        # Set the repo path to point to our folder
        repo.path = folder

        # Add in an environment.yml
        folder.joinpath("environment.yml").touch()

        # Should now return True
        assert repo.is_conda() is True

        # Now remove environment.yml
        folder.joinpath("environment.yml").unlink()

        # Should now return False
        assert repo.is_conda() is False


def test_is_pep517(
    mocker: MockerFixture, temp_config_file, repo_folder_with_random_existing_files
):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Return True for repo.exists_local
        mocker.patch("pytoil.repo.Repo.exists_local", autospec=True, return_value=True)

        folder: pathlib.Path = repo_folder_with_random_existing_files

        repo = Repo(name="myrepo")

        # Set the repo path to point to our folder
        repo.path = folder

        # Add in an pyproject.toml
        folder.joinpath("pyproject.toml").touch()

        # Should now return True
        assert repo.is_pep517() is True

        # Now remove pyproject.toml
        folder.joinpath("pyproject.toml").unlink()

        # Should now return False
        assert repo.is_pep517() is False


@pytest.mark.parametrize(
    "repo_name, description, created_at, updated_at, size, license_name, exist_local,"
    + " exist_remote",
    [
        (
            "repo1",
            "my project",
            "2020-02-27",
            "2020-04-02",
            4096,
            "MIT License",
            True,
            True,
        ),
        (
            "repo2",
            "someguys project",
            "2021-01-18",
            "2021-01-23",
            1024,
            "Apache 2.0",
            False,
            True,
        ),
        (
            "repo3",
            "somegirls project",
            "2020-07-01",
            "2021-02-28",
            2048,
            "GPL v3",
            True,
            True,
        ),
    ],
)
def test_info_on_remote_repo(
    mocker: MockerFixture,
    temp_config_file,
    repo_name,
    description,
    created_at,
    updated_at,
    size,
    license_name,
    exist_remote,
    exist_local,
):
    """
    If a repo exists remotely, regardless of whether or not it is also
    local. info should return info from the API as it is more detailed.
    """

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Convince it the repo exists remotely and locally
        mocker.patch(
            "pytoil.repo.Repo.exists_remote", autospec=True, return_value=exist_remote
        )
        mocker.patch(
            "pytoil.repo.Repo.exists_local", autospec=True, return_value=exist_local
        )

        # Have the get_repo_info method just return our made up dict
        mocker.patch(
            "pytoil.api.API.get_repo_info",
            autospec=True,
            return_value={
                "name": repo_name,
                "description": description,
                "created_at": created_at,
                "updated_at": updated_at,
                "size": size,
                "license": license_name,
                "local": exist_local,
                "remote": exist_remote,
            },
        )

        repo = Repo(name=repo_name)

        assert repo.info() == {
            "name": repo_name,
            "description": description,
            "created_at": created_at,
            "updated_at": updated_at,
            "size": size,
            "license": license_name,
            "local": exist_local,
            "remote": exist_remote,
        }


@pytest.mark.parametrize(
    "repo_name, created_at, updated_at, size, exist_local, exist_remote",
    [
        (
            "repo1",
            "2021-03-01 10:43:19",
            "2021-03-01 10:48:19",
            4096,
            True,
            False,
        ),
        (
            "repo2",
            "2021-03-01 10:43:19",
            "2021-03-01 10:48:19",
            1024,
            True,
            False,
        ),
        (
            "repo3",
            "2021-03-01 10:43:19",
            "2021-03-01 10:48:19",
            2048,
            True,
            False,
        ),
    ],
)
def test_info_on_local_only_repo(
    mocker: MockerFixture,
    temp_config_file,
    repo_name,
    created_at,
    updated_at,
    size,
    exist_remote,
    exist_local,
):
    """
    If a repo exists locally only, we should instead get some info from Path.stat.

    This one is quite complicated because you can't mock out datetime
    so instead we:
    - patch out everything related to whether or not the repo exists
    - Create a fake dataclass `FakeStat`
    - This FakeStat houses the unix timestamps that datetime.strftime can operate on
    - In the parametrize we have the strftime outputs for those timestamps
    """

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Convince it the repo exists remotely and locally
        mocker.patch(
            "pytoil.repo.Repo.exists_remote", autospec=True, return_value=exist_remote
        )
        mocker.patch(
            "pytoil.repo.Repo.exists_local", autospec=True, return_value=exist_local
        )

        # Have the get_repo_info method just return our made up dict
        mocker.patch(
            "pytoil.api.API.get_repo_info",
            autospec=True,
            return_value={
                "name": repo_name,
                "created_at": created_at,
                "updated_at": updated_at,
                "size": size,
                "local": exist_local,
                "remote": exist_remote,
            },
        )

        class FakeStat(NamedTuple):
            st_ctime = 1614595399
            st_mtime = 1614595699
            st_size = size

        mocker.patch(
            "pytoil.repo.repo.pathlib.Path.stat", autospec=True, return_value=FakeStat()
        )

        repo = Repo(name=repo_name)

        assert repo.info() == {
            "name": repo_name,
            "created_at": created_at,
            "updated_at": updated_at,
            "size": size,
            "local": exist_local,
            "remote": exist_remote,
        }


def test_repo_info_raises_if_doesnt_exist_locally_or_remotely(
    mocker: MockerFixture, temp_config_file
):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        # Convince it the repo does not exist remotely or locally
        mocker.patch(
            "pytoil.repo.Repo.exists_remote", autospec=True, return_value=False
        )
        mocker.patch("pytoil.repo.Repo.exists_local", autospec=True, return_value=False)

        with pytest.raises(RepoNotFoundError):
            repo = Repo(name="blah")
            repo.info()


def test_repo_dispatch_env_correctly_identifies_conda(
    mocker: MockerFixture, temp_config_file
):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        mocker.patch("pytoil.repo.Repo.is_conda", autospec=True, return_value=True)

        repo = Repo(name="test")

        env = repo.dispatch_env()

        assert isinstance(env, CondaEnv)


def test_repo_dispatch_env_correctly_identifies_virtualenv(
    mocker: MockerFixture, temp_config_file
):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        mocker.patch("pytoil.repo.Repo.is_conda", autospec=True, return_value=False)

        mocker.patch("pytoil.repo.Repo.is_setuptools", autospec=True, return_value=True)

        repo = Repo(name="test")

        env = repo.dispatch_env()

        assert isinstance(env, VirtualEnv)


def test_repo_dispatch_env_returns_none_when_it_cant_detect(
    mocker: MockerFixture, temp_config_file
):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        mocker.patch("pytoil.repo.Repo.is_conda", autospec=True, return_value=False)

        mocker.patch(
            "pytoil.repo.Repo.is_setuptools", autospec=True, return_value=False
        )

        repo = Repo(name="test")

        env = repo.dispatch_env()

        assert env is None



"""
Tests for the API module.

Author: Tom Fleet
Created: 05/02/2021
"""


import httpx
import pytest
from pytest_httpx import HTTPXMock
from pytest_mock import MockerFixture

import pytoil
from pytoil.api import API


def test_api_init_passed():

    api = API(token="definitelynotatoken", username="me")

    assert api.token == "definitelynotatoken"
    assert api.username == "me"
    assert api.baseurl == "https://api.github.com/"
    assert api.headers == {
        "Accept": "application/vnd.github.v3+json",
        "Authorization": "token definitelynotatoken",
    }


def test_api_init_default(mocker: MockerFixture, temp_config_file):

    # Patch out the default CONFIG_PATH for our temp file
    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        api = API()

        assert api.token == "tempfiletoken"
        assert api.username == "tempfileuser"
        assert api.baseurl == "https://api.github.com/"
        assert api.headers == {
            "Accept": "application/vnd.github.v3+json",
            "Authorization": "token tempfiletoken",
        }


def test_api_repr_passed():

    api = API(token="definitelynotatoken", username="me")

    assert api.__repr__() == "API(token='definitelynotatoken', username='me')"


def test_api_repr_default(mocker: MockerFixture, temp_config_file):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        api = API()

        assert api.__repr__() == "API(token='tempfiletoken', username='tempfileuser')"


@pytest.mark.parametrize("bad_status_code", [400, 401, 403, 404, 500, 504, 505])
def test_get_raises_on_bad_status(
    httpx_mock: HTTPXMock, bad_status_code, mocker: MockerFixture, temp_config_file
):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        httpx_mock.add_response(
            url="https://api.github.com/user/repos", status_code=bad_status_code
        )

        api = API(token="definitelynotatoken", username="me")

        with pytest.raises(httpx.HTTPStatusError):
            api.get("user/repos")


def test_get_returns_correct_response(
    httpx_mock: HTTPXMock, fake_api_response, mocker: MockerFixture, temp_config_file
):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        httpx_mock.add_response(
            url="https://api.github.com/user/repos",
            json=fake_api_response,
            status_code=200,
        )

        api = API(token="definitelynotatoken", username="me")

        r = api.get("user/repos")

        assert r == fake_api_response


def test_get_user_repo_correctly_calls_get(mocker: MockerFixture, fake_api_response):

    mocker.patch("pytoil.api.API.get", autospec=True, return_value=fake_api_response)

    api = API(token="definitelynotatoken", username="me")

    assert api.get_repo(repo="fakey") == fake_api_response


def test_get_user_repos_correctly_calls_get(mocker: MockerFixture, fake_api_response):

    mocker.patch("pytoil.api.API.get", autospec=True, return_value=fake_api_response)

    api = API(token="definitelynotatoken", username="me")

    assert api.get_repos() == fake_api_response


def test_get_repo_names_correctly_calls_get_repos(
    mocker: MockerFixture, fake_api_response
):

    mocker.patch(
        "pytoil.api.API.get_repos", autospec=True, return_value=fake_api_response
    )

    api = API(token="definitelynotatoken", username="me")

    # The names are contained in the fake_api_response fixture in conftest.py
    assert api.get_repo_names() == ["repo1", "repo2", "repo3"]


@pytest.mark.parametrize(
    "repo_name, description, created_at, updated_at, size, license_dict",
    [
        (
            "repo1",
            "my project",
            "2020-02-27",
            "2020-04-02",
            4096,
            {"id": "some_id", "name": "MIT License"},
        ),
        (
            "repo2",
            "someguys project",
            "2021-01-18",
            "2021-01-23",
            1024,
            {"id": "some_id", "name": "Apache 2.0"},
        ),
        (
            "repo3",
            "somegirls project",
            "2020-07-01",
            "2021-02-28",
            2048,
            {"id": "some_id", "name": "GPL v3"},
        ),
    ],
)
def test_get_repo_info_correctly_calls_get_repo(
    mocker: MockerFixture,
    repo_name,
    description,
    created_at,
    updated_at,
    size,
    license_dict,
):

    # Have the get_repo method just return our made up dict
    mocker.patch(
        "pytoil.api.API.get_repo",
        autospec=True,
        return_value={
            "name": repo_name,
            "description": description,
            "created_at": created_at,
            "updated_at": updated_at,
            "size": size,
            "license": license_dict,
        },
    )

    api = API(token="definitelynotatoken", username="me")

    assert api.get_repo_info(repo=repo_name) == {
        "name": repo_name,
        "description": description,
        "created_at": created_at,
        "updated_at": updated_at,
        "size": size,
        "license": license_dict["name"],
    }


def test_get_repo_info_correctly_handles_missing_license(mocker: MockerFixture):

    # Have the get_repo method just return a made up dict with None for license
    mocker.patch(
        "pytoil.api.API.get_repo",
        autospec=True,
        return_value={
            "name": "repo",
            "description": "This is only a test",
            "created_at": "2021-01-02",
            "updated_at": "2021-01-03",
            "size": 1024,
            "license": None,
        },
    )

    api = API(token="definitelynotatoken", username="me")

    assert api.get_repo_info(repo="repo")["license"] is None



"""
Tests for the project info CLI command.


Author: Tom Fleet
Created: 11/04/2021
"""

from typing import Dict, Union

from pytest_mock import MockerFixture
from typer.testing import CliRunner

from pytoil.cli.main import app
from pytoil.config import Config
from pytoil.exceptions import RepoNotFoundError

runner = CliRunner()


def test_info_displays_properly(mocker: MockerFixture, fake_projects_dir):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.main.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    return_dict: Dict[str, Union[str, int, bool]] = {
        "name": "dingle",
        "size": 2048,
        "remote": True,
        "local": True,
        "created_at": "12/02/20",
        "updated_at": "18/02/20",
    }

    # Patch out repo.info to return whatever dict we want
    mocker.patch("pytoil.cli.main.Repo.info", autospec=True, return_value=return_dict)

    result = runner.invoke(app, ["info", "myproject"])
    assert result.exit_code == 0
    assert "Info for: 'myproject'" in result.stdout

    for key, val in return_dict.items():
        assert f"{key}: {val}" in result.stdout


def test_info_handles_repo_not_found(mocker: MockerFixture, fake_projects_dir):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.main.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Make repo.info raise a RepoNotFoundError
    mocker.patch(
        "pytoil.cli.main.Repo.info",
        autospec=True,
        side_effect=[RepoNotFoundError("IT DOESNT EXIST YOU FOOL")],
    )

    # Now CLI should give nice message and abort
    result = runner.invoke(app, ["info", "someproject"])
    assert result.exit_code == 1
    assert "Project: 'someproject' not found locally or on GitHub." in result.stdout
    assert "Aborted!" in result.stdout



"""
Module responsible for querying the GitHub RESTv3 API.

Author: Tom Fleet
Created: 04/02/2021
"""

from typing import Any, Dict, List, Optional, Union

import httpx

from pytoil.config import Config


class API:
    def __init__(
        self, token: Optional[str] = None, username: Optional[str] = None
    ) -> None:
        """
        Representation of the GitHub API.

        Args:
            token (str, optional): GitHub Personal Access Token.
                Defaults to value from config file.

            username (str, optional): Users GitHub username.
                Defaults to value from config file.
        """
        # If token passed, set it
        # if not, get from config
        self.token = token or Config.get().token
        # Same with username
        self.username = username or Config.get().username

        # NOTE: we might support gitlab or others in future
        # PR's welcome!
        self.baseurl: str = "https://api.github.com/"

        self.headers: Dict[str, str] = {
            "Accept": "application/vnd.github.v3+json",
            "Authorization": f"token {self.token}",
        }

    def __repr__(self) -> str:
        return (
            self.__class__.__qualname__
            + f"(token={self.token!r}, "
            + f"username={self.username!r})"
        )

    def get(self, endpoint: str) -> Dict[str, Any]:
        """
        Makes an authenticated request to a GitHub API endpoint.

        Generic base for more specific get methods below.

        Args:
            endpoint (str): Valid GitHub API endpoint.
                e.g. 'users/repos'.

        Raises:
            HTTPStatusError: If any HTTP error occurs, will raise an exception
                and give a description and standard HTTP status code.

        Returns:
            ApiResponse: JSON API response.
        """
        # This needs the token from the config
        # it could be invalid, we haven't checked until now
        Config.get().validate()

        r = httpx.get(url=self.baseurl + endpoint, headers=self.headers)
        r.raise_for_status()
        response: Dict[str, Any] = r.json()

        return response

    def get_repo(self, repo: str) -> Dict[str, Any]:
        """
        Hits the GitHub REST API 'repos/{owner}/{repo}' endpoint
        and parses the response.

        In our case `owner` is `username`.

        Args:
            repo (str): The name of the repo to fetch JSON for.

        Returns:
            Dict[str, Any]: JSON response for a particular repo.
        """

        return self.get(f"repos/{self.username}/{repo}")

    def get_repos(self) -> Dict[str, Any]:
        """
        Hits the GitHub REST API 'user/{username}/repos' endpoint and parses
        the response.

        Function similar to `get_repo` the difference being `get_repos` returns
        a list of JSON blobs each representing a repo belonging to `self.username`

        This endpoint requires no parameters because it is the
        'get repos for authenticated user' endpoint and since at this point we have
        `self.token` this automatically fills in the {username} for us.

        Returns:
            Dict[str, Any]: JSON response for a list of all users repos.
        """

        return self.get("user/repos")

    def get_repo_names(self) -> List[str]:
        """
        Hits the GitHub REST API 'user/{username}/repos' endpoint, parses
        the response, extracts the name of each repo and returns
        a list of these names.

        Returns:
            List[str]: List of user's repo names.
        """

        return [repo["name"] for repo in self.get_repos()]  # type: ignore
        # We type ignore here because satisfying
        # mypy when dealing with JSON is hard

    def get_repo_info(self, repo: str) -> Dict[str, Union[str, int]]:
        """
        Returns a dictionary of key information about a particular repo.

        Info Keys:
        - name
        - description
        - created_at
        - updated_at
        - size
        - license

        Args:
            repo (str): Name of the repo to get info for.

        Returns:
            Dict[str, Union[str, int]]: Dict of repo information.
        """

        raw_repo_data = self.get_repo(repo=repo)

        keys_to_get: List[str] = [
            "name",
            "description",
            "created_at",
            "updated_at",
            "size",
            "license",
        ]

        display_dict: Dict[str, Union[str, int]] = {
            key: raw_repo_data.get(key, "Not found") for key in keys_to_get
        }

        # License is itself a dict
        # Couldn't be bothered doing some clever recursive thing for one key
        if raw_repo_data["license"]:
            display_dict["license"] = raw_repo_data.get("license", "Not Found").get(
                "name", "Not Found"
            )

        return display_dict



"""
Module managing Conda environments.

Author: Tom Fleet
Created: 07/03/2021
"""

from __future__ import annotations

import pathlib
import shutil
import subprocess
from typing import Dict, List, Optional, Union

import yaml

from pytoil.environments import BaseEnvironment
from pytoil.exceptions import (
    BadEnvironmentFileError,
    CondaNotInstalledError,
    UnknownCondaInstallationError,
    VirtualenvAlreadyExistsError,
    VirtualenvDoesNotExistError,
)


class CondaEnv(BaseEnvironment):
    def __init__(self, name: str, project_path: pathlib.Path) -> None:
        """
        Representation of a conda virtual environment.

        Conda environments are simpler in a lot of ways because
        it's all managed under the hood by the `conda` command.

        We don't need to worry about interpreters or paths etc.
        Just the environment name is enough to identify it.

        Args:
            name (str): The name of the conda environment.
            project_path (pathlib.Path): Path to the root dir
                of the associated project.
        """
        self.name = name
        self._project_path = project_path.resolve()

    def __repr__(self) -> str:
        return (
            self.__class__.__qualname__
            + f"(name={self.name!r}, "
            + f"project_path={self._project_path!r})"
        )

    @property
    def project_path(self) -> pathlib.Path:
        return self._project_path

    @property
    def executable(self) -> pathlib.Path:
        envs_dir = self.get_envs_dir()

        return envs_dir.joinpath(f"{self.name}/bin/python")

    def raise_for_conda(self) -> None:
        """
        Helper method which either raises if no `conda` found.
        Or does nothing if it is found.

        Because we end up relying on `conda` for most of the logic
        it makes sense to abstract this step.

        Raises:
            CondaNotInstalledError: If `conda` not found on $PATH
        """
        if not bool(shutil.which("conda")):
            raise CondaNotInstalledError(
                """`conda` executable not installed or not found on $PATH.
            Check your installation."""
            )

    def exists(self) -> bool:
        """
        Determines whether a conda environment called `name`
        exists on the current system by checking if a valid
        interpreter exists inside the environments directory
        under that name.

        Raises:
            CondaNotInstalledError: If `conda` not found on $PATH.

        Returns:
            bool: True if the environment exists on system, else False.
        """

        self.raise_for_conda()

        return self.executable.exists()

    def create(self, packages: Optional[List[str]] = None) -> None:
        """
        Creates the conda environment described by the instance.

        Only default package is `python=3` which will cause conda to choose
        it's default python3.

        If packages are specified, these will be inserted into the conda create
        command.

        Raises:
            VirtualenvAlreadyExistsError: If the conda environment already exists.
        """

        self.raise_for_conda()

        if self.exists():
            raise VirtualenvAlreadyExistsError(
                f"Conda env: {self.name} already exists."
            )

        cmd: List[str] = ["conda", "create", "-y", "--name", f"{self.name}", "python=3"]

        if packages:
            cmd.extend(packages)

        try:
            subprocess.run(cmd, check=True)
        except subprocess.CalledProcessError:
            raise

    @staticmethod
    def create_from_yml(project_path: pathlib.Path) -> CondaEnv:
        """
        Creates a conda environment from the `environment.yml`
        contained in the root `project_path`

        Args:
            fp (pathlib.Path): Filepath of the project root.

        Raises:
            FileNotFoundError: If the `environment.yml` file does not exist.
            BadEnvironmentFileError: If the `environment.yml` file is
                formatted incorrectly.
            VirtualenvAlreadyExistsError: If the conda environment described
                by the `environment.yml` file already exists on system.
        """

        # Ensure we have a resolved project root filepath
        resolved_project_path = project_path.resolve()
        yml_file = resolved_project_path.joinpath("environment.yml")

        try:
            with open(yml_file) as f:
                env_dict: Dict[str, Union[List[str], str]] = yaml.full_load(f)
        except FileNotFoundError:
            raise

        env_name = env_dict.get("name")

        if not isinstance(env_name, str):
            raise BadEnvironmentFileError(
                """The environment yml file has an invalid format.
                Cannot determine the value for key: `name`."""
            )
        env = CondaEnv(name=env_name, project_path=resolved_project_path)
        env.raise_for_conda()

        if env.exists():
            raise VirtualenvAlreadyExistsError(f"Conda env: {env.name} already exists.")
        else:
            try:
                subprocess.run(
                    [
                        "conda",
                        "env",
                        "create",
                        "-y",
                        "--file",
                        f"{yml_file}",
                    ],
                    check=True,
                )
            except subprocess.CalledProcessError:
                raise

        return env

    def export_yml(self) -> None:
        """
        Exports an environment.yml file for the conda environment
        described by the instance.

        Raises:
            VirtualenvDoesNotExistError: If the conda env does not exist,
                an environment file cannot be created.
        """

        self.raise_for_conda()

        if not self.exists():
            raise VirtualenvDoesNotExistError(
                f"""Conda env: {self.name} does not exist.
        Create it first before exporting the environment file."""
            )

        cmd: List[str] = [
            "conda",
            "env",
            "export",
            "--from-history",
            "--name",
            self.name,
        ]

        try:
            yml_out = subprocess.run(
                cmd, check=True, capture_output=True, encoding="utf-8"
            )
        except subprocess.CalledProcessError:
            raise
        else:
            yml_file = self.project_path.joinpath("environment.yml")
            with open(yml_file, "w") as f:
                f.write(yml_out.stdout)

    def install(
        self,
        packages: List[str],
    ) -> None:
        """
        Installs `packages` into the conda environment
        described by `name`.

        Packages are passed straight through to conda so any versioning
        syntax will work as expected.

        Args:
            packages (List[str]): List of packages to install.
                If only one package, still must be a list
                e.g. ["numpy"],

        Raises:
            VirtualenvDoesNotExistError: If the conda environment does not exist.
        """

        self.raise_for_conda()

        if not self.exists():
            raise VirtualenvDoesNotExistError(
                f"""Conda env: {self.name} does not exist.
        Create it first before installing packages."""
            )

        # Install into specified env
        cmd: List[str] = ["conda", "install", "--name", f"{self.name}", "-y"]

        # Add specified packages
        cmd.extend(packages)

        try:
            subprocess.run(cmd, check=True)
        except subprocess.CalledProcessError:
            raise

    @staticmethod
    def get_envs_dir() -> pathlib.Path:
        """
        Tries to detect which anaconda/miniconda installation
        the user has by naively checking if the environment
        storage directories exist.

        Raises:
            UnknownCondaInstallationError: If neither ~/anaconda3 or
                ~/miniconda3 exist or are not directories.

        Returns:
            pathlib.Path: The path to the users conda
                environment storage directory.
        """

        # As far as I'm aware, there are only 2 possible locations
        miniconda = pathlib.Path.home().joinpath("miniconda3/envs")
        anaconda = pathlib.Path.home().joinpath("anaconda3/envs")

        if miniconda.exists() and miniconda.is_dir():
            return miniconda
        elif anaconda.exists() and anaconda.is_dir():
            return anaconda
        else:
            raise UnknownCondaInstallationError(
                "Could not autodetect the conda environments directory."
            )



"""
Tests for the show CLI command.

Author: Tom Fleet
Created: 11/04/2021
"""

from typing import List

from pytest_mock import MockerFixture
from typer.testing import CliRunner

import pytoil
from pytoil.cli.main import app
from pytoil.config import Config

runner = CliRunner()


def test_show_local_returns_correct_directory_names(
    mocker: MockerFixture, fake_projects_dir
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.show.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    result = runner.invoke(app, ["show", "local"])
    assert result.exit_code == 0
    assert "Local projects" in result.stdout

    assert "project1" in result.stdout
    assert "myproject" in result.stdout
    assert "dingleproject" in result.stdout
    assert "anotherone" in result.stdout

    assert ".ishouldnt_show_up" not in result.stdout


def test_show_local_returns_special_message_if_no_local_projects(
    mocker: MockerFixture, empty_projects_dir
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=empty_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.show.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    result = runner.invoke(app, ["show", "local"])
    assert result.exit_code == 0

    assert "You don't have any local projects yet!" in result.stdout


def test_show_remote_returns_correct_project_names(
    mocker: MockerFixture, temp_config_file
):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        fake_repo_names: List[str] = ["repo1", "hellorepo", "myrepo", "anotherrepo"]

        mocker.patch(
            "pytoil.api.api.API.get_repo_names",
            autospec=True,
            return_value=fake_repo_names,
        )

        result = runner.invoke(app, ["show", "remote"])
        assert result.exit_code == 0
        assert "Remote projects" in result.stdout

        for repo in fake_repo_names:
            assert repo in result.stdout


def test_show_remote_returns_special_message_if_no_remote_projects(
    mocker: MockerFixture, temp_config_file
):

    with mocker.patch.object(pytoil.config.config, "CONFIG_PATH", temp_config_file):

        fake_repo_names: List[str] = []

        mocker.patch(
            "pytoil.api.api.API.get_repo_names",
            autospec=True,
            return_value=fake_repo_names,
        )

        result = runner.invoke(app, ["show", "remote"])
        assert result.exit_code == 0
        assert "You don't have any remote projects yet!" in result.stdout


def test_show_all_returns_both_correctly(mocker: MockerFixture, fake_projects_dir):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.show.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    fake_repo_names: List[str] = ["repo1", "hellorepo", "myrepo", "anotherrepo"]

    mocker.patch(
        "pytoil.api.api.API.get_repo_names",
        autospec=True,
        return_value=fake_repo_names,
    )

    result = runner.invoke(app, ["show", "all"])
    assert result.exit_code == 0

    assert "Local projects" in result.stdout
    assert "Remote projects" in result.stdout

    assert "project1" in result.stdout
    assert "myproject" in result.stdout
    assert "dingleproject" in result.stdout
    assert "anotherone" in result.stdout

    assert ".ishouldnt_show_up" not in result.stdout

    for repo in fake_repo_names:
        assert repo in result.stdout


def test_show_all_returns_special_message_if_no_local_projects(
    mocker: MockerFixture, empty_projects_dir
):

    # Here we have remotes but no locals
    fake_repo_names: List[str] = ["repo1", "hellorepo", "myrepo", "anotherrepo"]

    mocker.patch(
        "pytoil.api.api.API.get_repo_names",
        autospec=True,
        return_value=fake_repo_names,
    )

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=empty_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.show.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    result = runner.invoke(app, ["show", "all"])
    assert result.exit_code == 0

    assert "You don't have any local projects yet!" in result.stdout

    # Should still show remotes
    assert "Remote projects" in result.stdout

    for repo in fake_repo_names:
        assert repo in result.stdout


def test_show_all_returns_special_message_if_no_remote_projects(
    mocker: MockerFixture, fake_projects_dir
):

    # Here we have locals but no remotes
    fake_repo_names: List[str] = []

    mocker.patch(
        "pytoil.api.api.API.get_repo_names",
        autospec=True,
        return_value=fake_repo_names,
    )

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.show.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    result = runner.invoke(app, ["show", "all"])
    assert result.exit_code == 0

    assert "You don't have any remote projects yet!" in result.stdout

    # Should still show locals
    assert "Local projects" in result.stdout

    assert "project1" in result.stdout
    assert "myproject" in result.stdout
    assert "dingleproject" in result.stdout
    assert "anotherone" in result.stdout

    assert ".ishouldnt_show_up" not in result.stdout


def test_show_diff_shows_what_wed_expect(mocker: MockerFixture, fake_projects_dir):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.show.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # This time add some locals in so we can work out the diff
    fake_repo_names: List[str] = [
        "repo1",
        "hellorepo",
        "myrepo",
        "anotherrepo",
        "project1",
        "myproject",
        "dingleproject",
        "anotherone",
    ]

    mocker.patch(
        "pytoil.api.api.API.get_repo_names",
        autospec=True,
        return_value=fake_repo_names,
    )

    # Now the diff is just the repos
    actual_diff: List[str] = [
        "repo1",
        "hellorepo",
        "myrepo",
        "anotherrepo",
    ]

    result = runner.invoke(app, ["show", "diff"])
    assert result.exit_code == 0
    assert "Remote projects that are not local" in result.stdout

    for project in actual_diff:
        assert project in result.stdout


def test_show_diff_shows_special_message_if_local_and_remote_are_synced(
    mocker: MockerFixture, fake_projects_dir
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=True,
    )

    mocker.patch(
        "pytoil.cli.show.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Now our fake repo names exactly match the set of local projects
    fake_repo_names: List[str] = [
        "project1",
        "myproject",
        "dingleproject",
        "anotherone",
    ]

    mocker.patch(
        "pytoil.api.api.API.get_repo_names",
        autospec=True,
        return_value=fake_repo_names,
    )

    result = runner.invoke(app, ["show", "diff"])
    assert result.exit_code == 0
    assert (
        "Your local and remote projects are all synced up. Nothing to show!"
        in result.stdout
    )



"""
Module responsible for handling user config.

Author: Tom Fleet
Created: 05/02/2021
"""

from __future__ import annotations

import pathlib
from dataclasses import dataclass
from typing import List, Optional, TypedDict

import yaml

from pytoil.exceptions import InvalidConfigError

# Default value for projects_dir
DEFAULT_PROJECTS_DIR = pathlib.Path.home().joinpath("Development").resolve()
# Default config path location
CONFIG_PATH = pathlib.Path.home().joinpath(".pytoil.yml").resolve()


class ConfigDict(TypedDict):
    """
    Type checking for the config dictionary.

    Effectively a config schema, same as the dataclass
    with the exception of pathlib.Paths which are here
    represented as a string as this is how they will be
    stored in the config file and thus how they will be brought
    into python by pyyaml.
    """

    username: str
    token: str
    projects_dir: str
    vscode: bool
    common_packages: Optional[List[str]]


@dataclass
class Config:
    """
    Representation of the pytoil config.

    Only real usage is the `.get` classmethod.
    Used as the global configuration management class for the
    whole project.


    Args:
        username (str): Users GitHub username.
            Defaults to "UNSET".
        token (str): Users GitHub personal access token.
            Defaults to "UNSET".
        projects_dir (pathlib.Path): Directory in which user stores
            their development projects. Defaults to ~/Development.
        vscode (bool): Whether or not the user uses vscode as their editor.
            Defaults to False.
    """

    username: str = "UNSET"
    token: str = "UNSET"
    projects_dir: pathlib.Path = DEFAULT_PROJECTS_DIR
    vscode: bool = False
    common_packages: Optional[List[str]] = None

    def validate(self) -> None:
        """
        Helper method to validate the config.

        If either `username` or `token` are the fallback value
        "UNSET" this will raise an InvalidConfigError.

        Call before anything where these are required.
        """

        if self.username == "UNSET":
            raise InvalidConfigError(
                """No GitHub username set in ~/.pytoil.yml.
        Please set your GitHub username in the config file with the key: `username`."""
            )
        elif self.token == "UNSET":
            raise InvalidConfigError(
                """No GitHub personal access token set in ~/.pytoil.yml.
        Please set your GitHub personal access token in the config file with
        the key: `token`."""
            )
        else:
            return None

    @classmethod
    def get(cls) -> Config:
        """
        Fetches the configuration from the ~/.pytoil.yml config
        file and returns a Config object with parameters set from
        the file.

        If a key is not present in the file, or if the value associated to
        that key is blank. The default value for that key will be used.

        If the file does not exist, a FileNotFoundError will be raised.

        Raises:
            FileNotFoundError: If config file `~/.pytoil.yml` does not exist.

        Returns:
            Config: Config object with parameters parsed from the file.
        """

        try:
            with open(CONFIG_PATH, mode="r", encoding="utf-8") as f:
                config_dict = yaml.full_load(f)
        except FileNotFoundError:
            raise
        else:
            # Get the config from unpacking the dict
            config = Config(
                username=config_dict.get("username", "UNSET"),
                token=config_dict.get("token", "UNSET"),
                projects_dir=pathlib.Path(
                    config_dict.get("projects_dir", str(DEFAULT_PROJECTS_DIR))
                ),
                vscode=config_dict.get("vscode", False),
                common_packages=config_dict.get("common_packages", None),
            )

            return config

    def to_dict(self) -> ConfigDict:
        """
        Generates a properly formatted dictionary of the current
        config.

        Returns:
            ConfigDict: Dictionary showing the current config state.
        """

        config_dict: ConfigDict = {
            "username": self.username,
            "token": self.token,
            "projects_dir": str(self.projects_dir),
            "vscode": self.vscode,
            "common_packages": self.common_packages,
        }

        return config_dict

    def show(self) -> None:
        """
        Pretty prints the current config.
        """

        for key, val in self.to_dict().items():
            print(f"{key}: {val!r}")

    def write(self) -> None:
        """
        Overwrites the config file with the attributes
        from the calling instance.
        """

        config_dict = self.to_dict()

        with open(CONFIG_PATH, mode="w", encoding="utf-8") as f:
            yaml.dump(config_dict, f)



"""
Tests for the VSCode opening and settings
logic.

Author: Tom Fleet
Created: 06/03/2021
"""

import json
import pathlib
import subprocess
from typing import Any, Dict

import pytest
from pytest_mock import MockerFixture

from pytoil.exceptions import CodeNotInstalledError
from pytoil.vscode import VSCode


def test_vscode_init():

    fake_project = pathlib.Path("fakeproject")

    code = VSCode(root=fake_project)

    assert code.root == fake_project.resolve()


def test_vscode_repr():

    fake_project = pathlib.Path("fakeproject")

    code = VSCode(root=fake_project)

    assert code.__repr__() == f"VSCode(root={fake_project.resolve()!r})"


def test_vscode_raise_if_not_installed_works_on_missing_code(mocker: MockerFixture):

    # Mock out the return from shutil.which to be False
    mocker.patch("pytoil.vscode.vscode.shutil.which", autospec=True, return_value=False)

    fake_project = pathlib.Path("/Users/me/project1")

    with pytest.raises(CodeNotInstalledError):
        code = VSCode(root=fake_project)
        code.raise_if_not_installed()


def test_vscode_raise_if_not_installed_does_nothing_when_code_installed(
    mocker: MockerFixture,
):

    # Mock out the return from shutil.which to be True
    mocker.patch("pytoil.vscode.vscode.shutil.which", autospec=True, return_value=True)

    fake_project = pathlib.Path("/Users/me/project1")

    # If this raises, the test fails
    code = VSCode(root=fake_project)
    code.raise_if_not_installed()


def test_vscode_open_passes_correct_command_to_subprocess(mocker: MockerFixture):

    # Mock out the raise_if_not_installed call
    mocker.patch(
        "pytoil.vscode.VSCode.raise_if_not_installed", autospec=True, return_value=None
    )

    mock_subprocess = mocker.patch("pytoil.vscode.vscode.subprocess.run", autospec=True)

    fake_project = pathlib.Path("/Users/me/project1")

    code = VSCode(root=fake_project)

    code.open()

    mock_subprocess.assert_called_once_with(
        ["code", f"{fake_project.resolve()}"], check=True
    )


def test_vscode_open_raises_on_subprocess_error(mocker: MockerFixture):

    # Mock out the raise_if_not_installed call
    mocker.patch(
        "pytoil.vscode.VSCode.raise_if_not_installed", autospec=True, return_value=None
    )

    mocker.patch(
        "pytoil.vscode.vscode.subprocess.run",
        autospec=True,
        side_effect=[subprocess.CalledProcessError(-1, "cmd")],
    )

    fake_project = pathlib.Path("/Users/me/project1")

    code = VSCode(root=fake_project)

    with pytest.raises(subprocess.CalledProcessError):
        code.open()


def test_vscode_set_python_path_works_on_non_existent_settings(
    mocker: MockerFixture, fake_projects_dir
):

    fake_project: pathlib.Path = fake_projects_dir.joinpath("myproject")
    settings = fake_project.joinpath(".vscode/settings.json")

    code = VSCode(root=fake_project)

    ppath = fake_project.joinpath("/usr/bin/sillypython")

    # Assert no settings before
    assert not settings.exists()

    # Set the python path
    code.set_python_path(ppath)

    # Now it should exist
    assert settings.exists()

    # Get contents
    with open(settings, mode="r", encoding="utf-8") as f:
        written_settings_dict: Dict[str, Any] = json.load(f)

    expected: Dict[str, Any] = {"python.pythonPath": f"{ppath.resolve()}"}

    assert written_settings_dict == expected


def test_vscode_set_python_path_works_on_existing_settings(
    mocker: MockerFixture, fake_projects_dir
):

    fake_project: pathlib.Path = fake_projects_dir.joinpath("myproject")
    settings = fake_project.joinpath(".vscode/settings.json")
    settings.parent.mkdir()
    settings.touch()

    code = VSCode(root=fake_project)

    ppath = fake_project.joinpath("/usr/bin/sillypython")

    settings_dict: Dict[str, Any] = {
        "editor.suggestSelection": True,
        "code-runner.runInTerminal": False,
        "python.linting.mypyEnabled": True,
        "python.linting.blackPath": "usr/bin/black",
        "randomSetting": "yes",
        "HowHardThisIsToComeUpWith": 10,
        "python.pythonPath": "/usr/bin/python",
        "python.testing.pytestEnabled": False,
    }

    # Write our fake settings to the right place
    with open(settings, mode="w", encoding="utf-8") as f:
        json.dump(settings_dict, f, sort_keys=True, indent=4)

    assert settings.exists()

    expected: Dict[str, Any] = {
        "editor.suggestSelection": True,
        "code-runner.runInTerminal": False,
        "python.linting.mypyEnabled": True,
        "python.linting.blackPath": "usr/bin/black",
        "randomSetting": "yes",
        "HowHardThisIsToComeUpWith": 10,
        "python.pythonPath": "/usr/bin/sillypython",
        "python.testing.pytestEnabled": False,
    }

    code.set_python_path(python_path=ppath)

    # Get new contents
    with open(settings, mode="r", encoding="utf-8") as f:
        written_settings_dict: Dict[str, Any] = json.load(f)

    assert written_settings_dict == expected


def test_vscode_set_python_path_works_on_empty_file(
    mocker: MockerFixture, fake_projects_dir
):

    fake_project: pathlib.Path = fake_projects_dir.joinpath("myproject")
    settings = fake_project.joinpath(".vscode/settings.json")
    settings.parent.mkdir()
    settings.touch()

    code = VSCode(root=fake_project)

    ppath = fake_project.joinpath("/usr/bin/sillypython")

    assert settings.exists()

    # Assert it's empty
    assert len(settings.read_text()) == 0

    code.set_python_path(python_path=ppath)

    assert len(settings.read_text()) != 0

    # Get new contents
    with open(settings, mode="r", encoding="utf-8") as f:
        written_settings_dict: Dict[str, Any] = json.load(f)

    expected: Dict[str, Any] = {"python.pythonPath": f"{ppath.resolve()}"}

    assert written_settings_dict == expected



"""
CLI show command.

Author: Tom Fleet
Created: 28/02/2021
"""

from typing import List, Set

import typer

from pytoil.api import API
from pytoil.cli.utils import (
    get_local_project_list,
    get_local_project_set,
    get_remote_project_list,
    get_remote_project_set,
)
from pytoil.config import Config

app = typer.Typer(no_args_is_help=True)


# Callback for documentation only
@app.callback()
def show() -> None:
    """
    View your local/remote projects.
    """


@app.command()
def local() -> None:
    """
    Show all your local projects.
    """

    config = Config.get()
    config.validate()

    local_projects: List[str] = get_local_project_list(config.projects_dir)

    n_locals: int = len(local_projects)

    assert n_locals >= 0

    if n_locals > 0:

        typer.secho(
            f"\nLocal projects ({n_locals}):\n", fg=typer.colors.BLUE, bold=True
        )
        for project in local_projects:
            typer.echo(f"- {project}")
    else:
        typer.secho("You don't have any local projects yet!", fg=typer.colors.YELLOW)


@app.command()
def remote() -> None:
    """
    Show all your remote projects.
    """

    config = Config.get()
    config.validate()

    api = API()

    remote_projects: List[str] = get_remote_project_list(api)

    n_remotes: int = len(remote_projects)

    assert n_remotes >= 0

    if n_remotes > 0:

        typer.secho(
            f"\nRemote projects ({n_remotes}):\n", fg=typer.colors.BLUE, bold=True
        )
        for project in remote_projects:
            typer.echo(f"- {project}")
    else:
        typer.secho("You don't have any remote projects yet!", fg=typer.colors.YELLOW)


# Can't call it all, keyword!
@app.command(name="all")
def all_() -> None:
    """
    Show both local and remote projects.
    """

    config = Config.get()
    config.validate()

    api = API()

    local_projects: List[str] = get_local_project_list(config.projects_dir)

    remote_projects: List[str] = get_remote_project_list(api)

    n_locals: int = len(local_projects)
    n_remotes: int = len(remote_projects)

    assert n_locals >= 0
    assert n_remotes >= 0

    # Show locals first
    if n_locals > 0:
        typer.secho(
            f"\nLocal projects ({n_locals}):\n", fg=typer.colors.BLUE, bold=True
        )
        for project in local_projects:
            typer.echo(f"- {project}")
    else:
        typer.secho("You don't have any local projects yet!", fg=typer.colors.YELLOW)

    # Now remotes
    if n_remotes > 0:
        typer.secho(
            f"\nRemote projects ({n_remotes}):\n", fg=typer.colors.BLUE, bold=True
        )
        for project in remote_projects:
            typer.echo(f"- {project}")
    else:
        typer.secho("You don't have any remote projects yet!", fg=typer.colors.YELLOW)


@app.command()
def diff() -> None:
    """
    Show the difference in local/remote projects.

    i.e. those projects you have remotely but not locally.
    """

    # Everything below requires a valid config
    config = Config.get()
    config.validate()

    api = API()

    local_projects: Set[str] = get_local_project_set(config.projects_dir)

    remote_projects: Set[str] = get_remote_project_set(api)
    difference: Set[str] = remote_projects.difference(local_projects)

    n_diff: int = len(difference)

    # Internal correctness
    assert n_diff >= 0

    if n_diff > 0:

        typer.secho(
            f"\nRemote projects that are not local ({n_diff}):\n",
            fg=typer.colors.BLUE,
            bold=True,
        )
        for project in sorted(list(difference), key=str.casefold):
            typer.echo(f"- {project}")
    else:

        typer.secho(
            "Your local and remote projects are all synced up. Nothing to show!",
            fg=typer.colors.GREEN,
        )



"""
Module responsible for managing local and remote
repos (projects).

Author: Tom Fleet
Created: 05/02/2021
"""

from __future__ import annotations

import pathlib
import re
import shutil
import subprocess
from datetime import datetime
from typing import Dict, Optional, Set, Union

from pytoil.api import API
from pytoil.config import Config
from pytoil.environments import BaseEnvironment, CondaEnv, VirtualEnv
from pytoil.exceptions import (
    GitNotInstalledError,
    InvalidRepoPathError,
    InvalidURLError,
    LocalRepoExistsError,
    RepoNotFoundError,
)

# Stupidly basic regex, I'm bad at these
REPO_REGEX = re.compile(r"https://github.com/[\w]+/[\w]+.git")
REPO_PATH_REGEX = re.compile(r"[\w]+/[\w]+")


class Repo:
    def __init__(self, name: str, owner: Optional[str] = None) -> None:
        """
        Representation of a Git/GitHub repo.

        The GitHub url is constructed from `owner` and `name` and is accesible
        through the `.url` read-only property.

        If the repo has been cloned and/or exists locally, the `.path` property
        will be set to a pathlib.Path pointing to the root of the cloned repo.

        In this sense, the `Repo` object can represent both a remote GitHub
        repo and a local project.

        Args:
            name (str): The name of the GitHub repo.

            owner (Optional[str], optional): The owner of the GitHub repo.
                Defaults to value `username` from config file.
        """
        self.owner = owner or Config.get().username
        self.name = name

        self.url: str = f"https://github.com/{self.owner}/{self.name}.git"

        # The path this repo would have were it to be cloned locally
        self._path: pathlib.Path = Config.get().projects_dir.joinpath(self.name)

    def __repr__(self) -> str:
        return (
            self.__class__.__qualname__ + f"(owner={self.owner!r}, name={self.name!r})"
        )

    @property
    def path(self) -> pathlib.Path:
        return self._path

    @path.setter
    def path(self, value: pathlib.Path) -> None:
        self._path = value

    @classmethod
    def from_url(cls, url: str) -> Repo:
        """
        Constructs a Repo by extracting `owner` and `name`
        from a valid GitHub Repo URL.

        Args:
            url (str): Valid GitHub Repo URL.

        Raises:
            InvalidURLError: If URL does not match valid REGEX.

        Returns:
            Repo: Repo with `owner` and `name` set from `url`.
        """

        if not REPO_REGEX.match(url):
            raise InvalidURLError(f"The URL: {url!r} is not a valid GitHub repo URL.")
        owner = url.rsplit(".git")[0].split("/")[-2]
        name = url.rsplit(".git")[0].split("/")[-1]

        return cls(name=name, owner=owner)

    @classmethod
    def from_path(cls, path: str) -> Repo:
        """
        Constructs a Repo by extracting `owner` and `name`
        from a valid shorthand repo path.

        Args:
            path (str): Valid shorthand path
                e.g. 'FollowTheProcess/pytoil'

        Raises:
            InvalidRepoPathError: If path does match valid REGEX.

        Returns:
            Repo: Repo with `owner` and `name` set from `path`.
        """

        if not REPO_PATH_REGEX.match(path):
            raise InvalidRepoPathError(
                f"The repo path: {path!r} is not a valid GitHub repo path."
            )
        owner, name = path.split("/")

        return cls(name=name, owner=owner)

    def exists_local(self) -> bool:
        """
        Determines whether or not the repo exists
        locally in configured projects folder.

        Returns:
            bool: True if repo exists locally, else False.
        """
        return self.path.exists()

    def exists_remote(self) -> bool:
        """
        Determines whether or not the repo called
        `self.name` exists under configured username.

        Returns:
            bool: True if repo exists on GitHub, else False.
        """

        api = API()
        repos: Set[str] = set(api.get_repo_names())

        return self.name in repos

    def clone(self) -> None:
        """
        Invokes git in a subprocess to clone the repo
        represented by the instance.

        Raises:
            GitNotInstalledError: If 'git' not found on $PATH.
            LocalRepoExistsError: If repo already exists in configured
                projects_dir.
            RepoNotFoundError: If repo not found on GitHub.
        """

        # Get the user config from file and validate
        config = Config.get()
        config.validate()

        # Check if git is installed
        if not bool(shutil.which("git")):
            raise GitNotInstalledError(
                """'git' executable not installed or not found on $PATH.
        Check your git installation."""
            )
        # Check if its already been cloned
        elif self.exists_local():
            raise LocalRepoExistsError(
                f"""The repo {self.name} already exists at {self.path}.
        Cannot clone a repo that already exists."""
            )
        # Check if the remote repo actually exists
        elif not self.exists_remote():
            raise RepoNotFoundError(f"Repo: {self.url} not found on GitHub")
        else:
            # If we get here, we can safely clone
            try:
                subprocess.run(
                    ["git", "clone", f"{self.url}"], check=True, cwd=config.projects_dir
                )
            except subprocess.CalledProcessError:
                raise
            else:
                # If clone succeeded, set self.path
                self.path = config.projects_dir.joinpath(self.name)

    def init(self) -> None:
        """
        Invokes git in a subprocess to initialise a new repo
        described by the instance.

        Raises:
            GitNotInstalledError: If git not on $PATH.
            LocalRepoExistsError: If repo already exists.
        """

        # Get the user config from file and validate
        config = Config.get()
        config.validate()

        # Check if git is installed
        if not bool(shutil.which("git")):
            raise GitNotInstalledError(
                """'git' executable not installed or not found on $PATH.
        Check your git installation."""
            )
        else:
            # If we get here, we can safely init
            try:
                subprocess.run(["git", "init"], check=True, cwd=self.path)
            except subprocess.CalledProcessError:
                raise

    def info(self) -> Dict[str, Union[str, int, bool]]:
        """
        Returns summary information about the repo
        from the API or Path.stat.

        Prefers the API info as it is more detailed.
        Will fall back to Path.stat only if the project
        is not available remotely.

        Returns:
            Dict[str, Union[str, int]]: Summary info.
        """

        info_dict: Dict[str, Union[str, int, bool]] = {}

        # Path.stat returns a UNIX timestamp for dates/times
        str_time_format: str = r"%Y-%m-%d %H:%M:%S"

        if self.exists_remote():
            # If remote, the API does all the work
            api = API()
            info_dict.update(api.get_repo_info(repo=self.name))
            info_dict["remote"] = True
            info_dict["local"] = self.exists_local()
        elif self.exists_local():
            # If local, we can get a bit of info from os.stat
            info_dict.update(
                {
                    "name": self.path.name,
                    "created_at": datetime.utcfromtimestamp(
                        self.path.stat().st_ctime
                    ).strftime(str_time_format),
                    "updated_at": datetime.utcfromtimestamp(
                        self.path.stat().st_mtime
                    ).strftime(str_time_format),
                    "size": self.path.stat().st_size,
                    "local": True,
                    "remote": self.exists_remote(),
                },
            )
        else:
            raise RepoNotFoundError(
                f"Repo: {self.name!r} does not exist locally or remotely."
            )

        return info_dict

    def _does_file_exist(self, file: str) -> bool:
        """
        Helper method to determine whether or not a particular file
        exists in the root of the local repo.

        Args:
            file (str): Name of the file to check for.

        Raises:
            RepoNotFoundError: If the repo does not exist locally.

        Returns:
            bool: True if file exists, else False.
        """

        if not self.exists_local():
            raise RepoNotFoundError(f"Repo: {self.path!r} not found locally.")
        else:
            return self.path.joinpath(file).exists()

    def is_setuptools(self) -> bool:
        """
        Is the project based on setuptools.

        i.e. does it contain a `setup.py` or a `setup.cfg`

        Returns:
            bool: True if project is setuptools, else False.
        """

        return self._does_file_exist("setup.cfg") or self._does_file_exist("setup.py")

    def is_conda(self) -> bool:
        """
        Is the project based on conda.

        i.e. does it contain an `environment.yml`

        Returns:
            bool: True if project is conda, else False.
        """

        return self._does_file_exist("environment.yml")

    def is_editable(self) -> bool:
        """
        Does the project support `pip install -e .`

        Must have a `setup.py` for this, `setup.cfg` on its own
        is not sufficient.

        Returns:
            bool: True if project is editable, else False.
        """

        return self._does_file_exist("setup.py")

    def is_pep517(self) -> bool:
        """
        Does the project comply with PEP517/518.

        i.e. does it have a `pyproject.toml`

        Returns:
            bool: True if project supports PEP517, else False.
        """

        return self._does_file_exist("pyproject.toml")

    def dispatch_env(self) -> Optional[BaseEnvironment]:
        """
        Returns the correct environment
        object for the calling `Repo` instance. Or `None` if it cannot
        detect the environment.

        Therefore all usage should first check for `None`

        Returns:
            Optional[BaseEnvironment]: Either the correct environment
                object if it was able to detect. Or None.
        """

        if self.is_conda():
            return CondaEnv(name=self.name, project_path=self.path)
        elif self.is_setuptools():
            return VirtualEnv(project_path=self.path)
        else:
            return None



"""
Collection of useful abstractions and internal
methods to be called in various CLI commands.

Author: Tom Fleet
Created: 08/03/2021
"""


import pathlib
from typing import List, Set

from pytoil.api import API


def get_local_project_list(projects_dir: pathlib.Path) -> List[str]:
    """
    Returns a sorted list of local projects.

    Args:
        projects_dir (pathlib.Path): Result of Config.projects_dir.

    Returns:
        List[str]: Sorted list of local project names.
    """

    local_projects: List[str] = sorted(
        [
            f.name
            for f in projects_dir.iterdir()
            if f.is_dir() and not f.name.startswith(".")
        ],
        key=str.casefold,
    )

    return local_projects


def get_local_project_set(projects_dir: pathlib.Path) -> Set[str]:
    """
    Returns a set of local project names.

    Args:
        projects_dir (pathlib.Path): Result of Config.projects_dir.

    Returns:
        Set[str]: Set of local project names.
    """

    local_projects: Set[str] = {
        f.name
        for f in projects_dir.iterdir()
        if f.is_dir() and not f.name.startswith(".")
    }

    return local_projects


def get_remote_project_list(api: API) -> List[str]:
    """
    Returns a sorted list of remote projects.

    Args:
        projects_dir (pathlib.Path): Result of Config.projects_dir.

    Returns:
        List[str]: Sorted list of remote project names.
    """

    remote_projects: List[str] = sorted(api.get_repo_names(), key=str.casefold)

    return remote_projects


def get_remote_project_set(api: API) -> Set[str]:
    """
    Returns a set of remote projects.

    Args:
        projects_dir (pathlib.Path): Result of Config.projects_dir.

    Returns:
        List[str]: Set of remote project names.
    """

    remote_projects: Set[str] = set(api.get_repo_names())

    return remote_projects



"""
Tests for the remove CLI command.

Author: Tom Fleet
Created: 11/04/2021
"""

from typing import List

import pytest
from pytest_mock import MockerFixture
from typer.testing import CliRunner

from pytoil.cli.main import app
from pytoil.config import Config

runner = CliRunner()


def test_remove_aborts_if_project_doesnt_exist(
    mocker: MockerFixture, fake_projects_dir
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=False,
    )

    mocker.patch(
        "pytoil.cli.main.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Try to remove a project thats not in our fake directory
    result = runner.invoke(app, ["remove", "i_was_never_here"])
    assert result.exit_code == 1
    assert "Project: 'i_was_never_here' not found in local filesystem" in result.stdout
    assert "Aborted!" in result.stdout


def test_remove_respects_no_input_on_confirmation(
    mocker: MockerFixture, fake_projects_dir
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=False,
    )

    mocker.patch(
        "pytoil.cli.main.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Make sure the path exists before
    assert fake_projects_dir.joinpath("myproject").exists()

    # Now we try and remove one that does exist, but without --force
    # simulate user input 'n'
    result = runner.invoke(app, ["remove", "myproject"], input="n\n")
    assert result.exit_code == 1
    assert "Aborted!" in result.stdout

    # Make sure it didn't actually delete it despite aborting
    assert fake_projects_dir.joinpath("myproject").exists()


def test_remove_respects_yes_input_on_confirmation(
    mocker: MockerFixture, fake_projects_dir
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=False,
    )

    mocker.patch(
        "pytoil.cli.main.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Make sure the path exists before
    assert fake_projects_dir.joinpath("myproject").exists()

    # Now we try and remove one that does exist, but without --force
    # simulate user input 'y'
    result = runner.invoke(app, ["remove", "myproject"], input="y\n")
    assert "Removing project: 'myproject'." in result.stdout
    assert "Done!" in result.stdout
    assert result.exit_code == 0

    # Make sure it performed the deletion
    assert not fake_projects_dir.joinpath("myproject").exists()


@pytest.mark.parametrize("force_flag", ["--force", "-f"])
def test_remove_respects_force_flag(
    mocker: MockerFixture, fake_projects_dir, force_flag
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=False,
    )

    mocker.patch(
        "pytoil.cli.main.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    # Make sure the path exists before
    assert fake_projects_dir.joinpath("myproject").exists()

    # Here we remove it but force deletion, shouldn't prompt to confirm
    result = runner.invoke(app, ["remove", "myproject", force_flag])
    assert (
        "This will remove 'myproject' from your local filesystem" not in result.stdout
    )
    assert "Are you sure?" not in result.stdout

    assert "Removing project: 'myproject'." in result.stdout
    assert "Done!" in result.stdout
    assert result.exit_code == 0

    # Make sure it performed the deletion
    assert not fake_projects_dir.joinpath("myproject").exists()


@pytest.mark.parametrize("force_flag", ["--force", "-f"])
def test_remove_works_on_multiple_args(
    mocker: MockerFixture, fake_projects_dir, force_flag
):

    fake_config = Config(
        username="test",
        token="testtoken",
        projects_dir=fake_projects_dir,
        vscode=False,
    )

    mocker.patch(
        "pytoil.cli.main.Config.get",
        autospec=True,
        return_value=fake_config,
    )

    projects_to_remove: List[str] = [
        "project1",
        "myproject",
        "dingleproject",
    ]

    for _ in projects_to_remove:
        assert fake_projects_dir.joinpath("myproject").exists()

    # Remove the list of projects
    result = runner.invoke(app, ["remove", *projects_to_remove, force_flag])
    assert "Are you sure?" not in result.stdout
    assert "Done!" in result.stdout
    assert result.exit_code == 0

    for project in projects_to_remove:
        assert (
            f"This will remove {project!r} from your local filesystem"
            not in result.stdout
        )
        assert f"Removing project: {project!r}." in result.stdout



"""
Abstract base class for all Virtual Environments

Author: Tom Fleet
Created: 07/03/2021
"""

import pathlib
from abc import ABC, abstractmethod
from typing import List, Optional


class BaseEnvironment(ABC):
    """
    Abstract Base Class providing an interface that
    all virtual environments must implement.
    """

    @property
    def project_path(self) -> pathlib.Path:
        """
        `.project_path` represents the root directory of the
        project associated to the virtual environment.
        """
        pass

    @property
    def executable(self) -> pathlib.Path:
        """
        `.executable` is the absolute path to the virtual environment's
        python interpreter e.g. "/Users/me/Development/project1/.venv/bin/python"
        """
        pass

    @abstractmethod
    def exists(self) -> bool:
        """
        `.exists()` checks whether the virtual environment
        exists. How it does this is up to the child class, but a
        good example would be simply checking if `self.executable` exists.
        """
        pass

    @abstractmethod
    def create(self, packages: Optional[List[str]] = None) -> None:
        """
        Method to create the virtual environment. If packages are specified
        these can be installed during environment creation.
        """
        pass

    @abstractmethod
    def install(
        self,
        packages: List[str],
    ) -> None:
        """
        Generic install method.

        Installs `packages` into the correct virtual environment.

        Args:
            packages (List[str]): List of valid packages to install.
        """
        pass
